from __future__ import absolute_import
from abc import ABCMeta
import os.path
import tempfile
from itertools import repeat
import shutil
import hashlib
import stat
import time
import logging
import errno
import json
from zipfile import ZipFile, BadZipfile
from collections import defaultdict
from nipype.interfaces.base import Directory, traits, isdefined
from arcana.dataset import Dataset, Field
from arcana.archive.base import (
    Archive, ArchiveSource, ArchiveSink, ArchiveSourceInputSpec,
    ArchiveSinkInputSpec, ArchiveSubjectSinkInputSpec,
    ArchiveVisitSinkInputSpec,
    ArchiveProjectSinkInputSpec,
    ArchiveSubjectSink, ArchiveVisitSink, ArchiveProjectSink,
    MULTIPLICITIES)
from arcana.archive.tree import Session, Subject, Project, Visit
from arcana.data_format import DataFormat
from arcana.utils import split_extension
from arcana.exception import (
    ArcanaError, ArcanaMissingDataException)
from arcana.utils import dir_modtime, NoContextWrapper
import re
import xnat  # NB: XnatPy not PyXNAT
from arcana.utils import PATH_SUFFIX, FIELD_SUFFIX
from .local import FIELDS_FNAME

logger = logging.getLogger('Arcana')

special_char_re = re.compile(r'[^a-zA-Z_0-9]')
tag_parse_re = re.compile(r'\((\d+),(\d+)\)')

RELEVANT_DICOM_TAG_TYPES = set(('UI', 'CS', 'DA', 'TM', 'SH', 'LO',
                                'PN', 'ST', 'AS'))

BUILTIN_XNAT_FIELDS = []


def lower(s):
    if s is None:
        return None
    return s.lower()


class XnatMixin(object):

    @property
    def session_id(self):
        return self.inputs.subject_id + '_' + self.inputs.visit_id


class XnatSourceInputSpec(ArchiveSourceInputSpec):
    project_id = traits.Str(mandatory=True, desc='The project ID')
    server = traits.Str(mandatory=True,
                        desc="The address of the XNAT server")
    user = traits.Str(
        mandatory=False,
        desc=("The XNAT username to connect with in with if not "
              "supplied it can be read from ~/.netrc (see "
              "https://xnat.readthedocs.io/en/latest/static/tutorial.html"
              "#connecting-to-a-server)"))
    password = traits.Password(
        mandatory=False,
        desc=("The XNAT password corresponding to the supplied username, if "
              "not supplied it can be read from ~/.netrc (see "
              "https://xnat.readthedocs.io/en/latest/static/tutorial.html"
              "#connecting-to-a-server)"))
    cache_dir = Directory(
        exists=True, desc=("Path to the base directory where the downloaded"
                           "datasets will be cached"))

    race_cond_delay = traits.Int(
        usedefault=True, default=30,
        desc=("The amount of time to wait before checking that the required "
              "dataset has been downloaded to cache by another process has "
              "completed if they are attempting to download the same dataset"))

    stagger = traits.Int(
        mandatory=False,
        desc=("Stagger the download of the required datasets by "
              "stagger_delay * subject_id seconds to avoid sending too many "
              "concurrent requests to XNAT"))


class XnatSource(ArchiveSource, XnatMixin):
    """
    A NiPype IO interface for grabbing datasets off DaRIS (analogous to
    DataGrabber)
    """

    input_spec = XnatSourceInputSpec

    def __init__(self, *args, **kwargs):
        self._check_md5 = kwargs.pop('check_md5', True)
        super(XnatSource, self).__init__(*args, **kwargs)

    @property
    def check_md5(self):
        return self._check_md5

    def _list_outputs(self):
        # FIXME: Should probably not prepend the project before this point
        subject_id = self.inputs.subject_id.split('_')[-1]
        visit_id = self.inputs.visit_id
        base_cache_dir = os.path.join(self.inputs.cache_dir,
                                      self.inputs.project_id)
        sess_kwargs = {}
        if isdefined(self.inputs.user):
            sess_kwargs['user'] = self.inputs.user
        if isdefined(self.inputs.password):
            sess_kwargs['password'] = self.inputs.password
        with xnat.connect(server=self.inputs.server,
                          **sess_kwargs) as xnat_login:
            project = xnat_login.projects[self.inputs.project_id]
            # Get primary session, derived and summary sessions and cache
            # dirs
            sessions = {}
            cache_dirs = {}
            for freq, derived in ([('per_session', False)] +
                                    zip(MULTIPLICITIES, repeat(True))):
                subj_label, sess_label = XnatArchive.get_labels(
                    freq, self.inputs.project_id, subject_id, visit_id)
                if freq == 'per_session' and derived:
                    sess_label += XnatArchive.PROCESSED_SUFFIX
                cache_dirs[(freq, derived)] = os.path.join(
                    base_cache_dir, subj_label, sess_label)
                try:
                    subject = project.subjects[subj_label]
                    sessions[(freq, derived)] = subject.experiments[
                        sess_label]
                except KeyError:
                    continue
            outputs = {}
            for dataset in self.datasets:
                try:
                    session = sessions[(dataset.frequency,
                                        dataset.derived)]
                except KeyError:
                    raise ArcanaMissingDataException(
                        "Did not find{} session for frequency '{}', "
                        "it was expected to find {} in"
                        .format(
                            (' derived' if dataset.frequency else ''),
                            dataset.frequency, dataset))
                cache_dir = cache_dirs[(dataset.frequency,
                                        dataset.derived)]
                try:
                    xdataset = session.scans[
                        dataset.basename(subject_id=subject_id,
                                         visit_id=visit_id)]
                except KeyError:
                    raise ArcanaError(
                        "Could not find '{}' dataset in session '{}' "
                        "(found {})".format(
                            dataset.prefixed_name, session.label,
                            "', '".join(session.scans.keys())))
                # Get filename
                fname = dataset.fname(subject_id=subject_id,
                                      visit_id=visit_id)
                # Get resource to check its MD5 digest
                xresource = self.get_resource(xdataset, dataset)
                need_to_download = True
                # FIXME: Should do a check to see if versions match
                if not os.path.exists(cache_dir):
                    os.makedirs(cache_dir)
                cache_path = os.path.join(cache_dir, fname)
                if os.path.exists(cache_path):
                    if self.check_md5:
                        try:
                            with open(cache_path +
                                      XnatArchive.MD5_SUFFIX) as f:
                                cached_digests = json.load(f)
                            digests = self.get_digests(xresource)
                            if cached_digests == digests:
                                need_to_download = False
                        except IOError:
                            pass
                    else:
                        need_to_download = False
                if need_to_download:
                    # The path to the directory which the files will be
                    # downloaded to.
                    tmp_dir = cache_path + '.download'
                    try:
                        # Attempt to make tmp download directory. This will
                        # fail if another process (or previous attempt) has
                        # already created it. In that case this process will
                        # wait to see if that download finishes successfully,
                        # and if so use the cached version.
                        os.mkdir(tmp_dir)
                    except OSError as e:
                        if e.errno == errno.EEXIST:
                            # Another process may be concurrently downloading
                            # the same file to the cache. Wait for
                            # 'race_cond_delay' seconds and then check that it
                            # has been completed or assume interrupted and
                            # redownload.
                            self.delayed_download(
                                tmp_dir, xresource, xdataset, dataset,
                                session.label, cache_path,
                                delay=self.inputs.race_cond_delay)
                        else:
                            raise
                    else:
                        self.download_dataset(
                            tmp_dir, xresource, xdataset, dataset,
                            session.label, cache_path)
                outputs[dataset.name + PATH_SUFFIX] = cache_path
            for field in self.fields:
                prefixed_name = field.prefixed_name
                session = sessions[(field.frequency,
                                    field.derived)]
                outputs[field.name + FIELD_SUFFIX] = field.dtype(
                    session.fields[prefixed_name])
        return outputs

    @classmethod
    def get_resource(cls, xdataset, dataset):
        try:
            xresource = xdataset.resources[
                dataset.format.xnat_resource_name]
        except KeyError:
            raise ArcanaError(
                "'{}' dataset is not available in '{}' format, "
                "available resources are '{}'"
                .format(
                    dataset.name, dataset.format.xnat_resource_name,
                    "', '".join(r.label
                                 for r in dataset.resources.values())))
        return xresource

    @classmethod
    def get_digests(cls, resource):
        """
        Downloads the MD5 digests associated with the files in a resource.
        These are saved with the downloaded files in the cache and used to
        check if the files have been updated on the server
        """
        result = resource.xnat_session.get(resource.uri + '/files')
        if result.status_code != 200:
            raise ArcanaError(
                "Could not download metadata for resource {}"
                .format(resource.id))
        return dict((r['Name'], r['digest'])
                    for r in result.json()['ResultSet']['Result'])

    @classmethod
    def download_dataset(cls, tmp_dir, xresource, xdataset, dataset,
                         session_label, cache_path):
        # Download resource to zip file
        zip_path = os.path.join(tmp_dir, 'download.zip')
        with open(zip_path, 'w') as f:
            xresource.xnat_session.download_stream(
                xresource.uri + '/files', f, format='zip', verbose=True)
        digests = cls.get_digests(xresource)
        # Extract downloaded zip file
        expanded_dir = os.path.join(tmp_dir, 'expanded')
        try:
            with ZipFile(zip_path) as zip_file:
                zip_file.extractall(expanded_dir)
        except BadZipfile as e:
            raise ArcanaError(
                "Could not unzip file '{}' ({})"
                .format(xresource.id, e))
        data_path = os.path.join(
            expanded_dir, session_label, 'scans',
            (xdataset.id + '-' + special_char_re.sub('_', xdataset.type)),
            'resources', dataset.format.xnat_resource_name, 'files')
        if not dataset.format.directory:
            # If the dataformat is not a directory (e.g. DICOM),
            # attempt to locate a single file within the resource
            # directory with the appropriate filename and add that
            # to be the complete data path.
            fnames = os.listdir(data_path)
            match_fnames = [
                f for f in fnames
                if (lower(split_extension(f)[-1]) ==
                    lower(dataset.format.extension))]
            if len(match_fnames) == 1:
                data_path = os.path.join(data_path, match_fnames[0])
            else:
                raise ArcanaMissingDataException(
                    "Did not find single file with extension '{}' "
                    "(found '{}') in resource '{}'"
                    .format(dataset.format.extension,
                            "', '".join(fnames), data_path))
        try:
            os.makedirs(os.path.dirname(cache_path))
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        shutil.move(data_path, cache_path)
        with open(cache_path + XnatArchive.MD5_SUFFIX, 'w') as f:
            json.dump(digests, f)
        shutil.rmtree(tmp_dir)

    @classmethod
    def delayed_download(cls, tmp_dir, xresource, xdataset, dataset,
                         session_label, cache_path, delay):
        logger.info("Waiting {} seconds for incomplete download of '{}' "
                    "initiated another process to finish"
                    .format(delay, cache_path))
        initial_mod_time = dir_modtime(tmp_dir)
        time.sleep(delay)
        if os.path.exists(cache_path):
            logger.info("The download of '{}' has completed "
                        "successfully in the other process, continuing"
                        .format(cache_path))
            return
        elif initial_mod_time != dir_modtime(tmp_dir):
            logger.info(
                "The download of '{}' hasn't completed yet, but it has"
                " been updated.  Waiting another {} seconds before "
                "checking again.".format(cache_path, delay))
            cls.delayed_download(tmp_dir, xresource, xdataset,
                                   dataset,
                                   session_label, cache_path, delay)
        else:
            logger.warning(
                "The download of '{}' hasn't updated in {} "
                "seconds, assuming that it was interrupted and "
                "restarting download".format(cache_path, delay))
            shutil.rmtree(tmp_dir)
            os.mkdir(tmp_dir)
            cls.download_dataset(
                tmp_dir, xresource, xdataset, dataset, session_label,
                cache_path)


class XnatSinkInputSpecMixin(object):
    project_id = traits.Str(mandatory=True, desc='The project ID')
    server = traits.Str('https://mf-erc.its.monash.edu.au', mandatory=True,
                        usedefault=True, desc="The address of the MF server")
    user = traits.Str(
        mandatory=False,
        desc=("The XNAT username to connect with in with if not "
              "supplied it can be read from ~/.netrc (see "
              "https://xnat.readthedocs.io/en/latest/static/tutorial.html"
              "#connecting-to-a-server)"))
    password = traits.Password(
        mandatory=False,
        desc=("The XNAT password corresponding to the supplied username, if "
              "not supplied it can be read from ~/.netrc (see "
              "https://xnat.readthedocs.io/en/latest/static/tutorial.html"
              "#connecting-to-a-server)"))
    cache_dir = Directory(
        exists=True, desc=("Path to the base directory where the downloaded"
                           "datasets will be cached"))


class XnatSinkInputSpec(ArchiveSinkInputSpec, XnatSinkInputSpecMixin):
    pass


class XnatSubjectSinkInputSpec(ArchiveSubjectSinkInputSpec,
                               XnatSinkInputSpecMixin):
    pass


class XnatVisitSinkInputSpec(ArchiveVisitSinkInputSpec,
                                 XnatSinkInputSpecMixin):
    pass


class XnatProjectSinkInputSpec(ArchiveProjectSinkInputSpec,
                               XnatSinkInputSpecMixin):
    pass


class XnatSinkMixin(XnatMixin):
    """
    A NiPype IO interface for putting derived datasets onto DaRIS (analogous
    to DataSink)
    """

    __metaclass__ = ABCMeta

    def _list_outputs(self):
        """Execute this module.
        """
        # Initiate output
        outputs = self._base_outputs()
        out_files = []
        missing_files = []
        # Open XNAT session
        sess_kwargs = {}
        if 'user' in self.inputs.trait_names():  # Because InputSpec is dynamic
            sess_kwargs['user'] = self.inputs.user
        if 'password' in self.inputs.trait_names():
            sess_kwargs['password'] = self.inputs.password
        logger.debug("Session kwargs: {}".format(sess_kwargs))
        with xnat.connect(server=self.inputs.server,
                          **sess_kwargs) as xnat_login:
            # Add session for derived scans if not present
            session, cache_dir = self._get_session(xnat_login)
            # Make session cache dir
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, stat.S_IRWXU | stat.S_IRWXG)
            # Loop through datasets connected to the sink and copy them to the
            # cache directory and upload to daris.
            for dataset in self.datasets:
                assert dataset.frequency == self.frequency
                assert dataset.derived, (
                    "{} (format: {}, freq: {}) isn't derived"
                    .format(dataset.name, dataset.format_name,
                            dataset.frequency))
                filename = getattr(self.inputs,
                                   dataset.name + PATH_SUFFIX)
                if not isdefined(filename):
                    missing_files.append(dataset.name)
                    continue  # skip the upload for this file
                ext = dataset.format.extension
                assert split_extension(filename)[1] == ext, (
                    "Mismatching extension '{}' for format '{}' ('{}')"
                    .format(split_extension(filename)[1],
                            dataset.format.name,
                            dataset.format.extension))
                src_path = os.path.abspath(filename)
                out_fname = dataset.fname()
                # Copy to local cache
                dst_path = os.path.join(cache_dir, out_fname)
                out_files.append(dst_path)
                shutil.copyfile(src_path, dst_path)
                # Create md5 digest
                with open(dst_path) as f:
                    digests = {out_fname: hashlib.md5(f.read()).hexdigest()}
                with open(dst_path + XnatArchive.MD5_SUFFIX, 'w') as f:
                    json.dump(digests, f)
                # Upload to XNAT
                xdataset = xnat_login.classes.MrScanData(
                    type=dataset.basename(), parent=session)
                # Delete existing resource
                # TODO: probably should have check to see if we want to
                #       override it
                try:
                    xresource = xdataset.resources[
                        dataset.format.name.upper()]
                    xresource.delete()
                except KeyError:
                    pass
                xresource = xdataset.create_resource(
                    dataset.format.name.upper())
                xresource.upload(dst_path, out_fname)
            for field in self.fields:
                assert field.frequency == self.frequency
                assert field.derived, ("{} isn't derived".format(
                    field))
                session.fields[field.prefixed_name] = getattr(
                    self.inputs, field.name + FIELD_SUFFIX)
        if missing_files:
            # FIXME: Not sure if this should be an exception or not,
            #        indicates a problem but stopping now would throw
            #        away the datasets that were created
            logger.warning(
                "Missing output datasets '{}' in XnatSink".format(
                    "', '".join(str(f) for f in missing_files)))
        # Return cache file paths
        outputs['out_files'] = out_files
        return outputs

    def _get_session(self, xnat_login):
        project = xnat_login.projects[self.inputs.project_id]
        # FIXME: Subject should probably be input without the project prefix
        try:
            subject_id = self.inputs.subject_id.split('_')[-1]
        except AttributeError:
            subject_id = None
        try:
            visit_id = self.inputs.visit_id
        except AttributeError:
            visit_id = None
        subj_label, sess_label = XnatArchive.get_labels(
            self.frequency, self.inputs.project_id, subject_id, visit_id)
        if self.frequency == 'per_session':
            sess_label += XnatArchive.PROCESSED_SUFFIX
            if visit_id is not None:
                visit_id += XnatArchive.PROCESSED_SUFFIX
        try:
            subject = project.subjects[subj_label]
        except KeyError:
            subject = xnat_login.classes.SubjectData(
                label=subj_label, parent=project)
        try:
            session = subject.experiments[sess_label]
        except KeyError:
            session = self._create_session(xnat_login, subj_label,
                                           sess_label)
        # Get cache dir for session
        cache_dir = os.path.abspath(os.path.join(
            self.inputs.cache_dir, self.inputs.project_id, subject.label,
            session.label))
        return session, cache_dir

    def _create_session(self, xnat_login, subject_id, visit_id):
        """
        This creates a derived session in a way that respects whether
        the acquired session has been shared into another project or not.

        If we weren't worried about this we could just use

            session = xnat_login.classes.MrSessionData(label=proc_session_id,
                                                       parent=subject)
        """
        uri = ('/data/archive/projects/{}/subjects/{}/experiments/{}'
               .format(self.inputs.project_id, subject_id, visit_id))
        query = {'xsiType': 'xnat:mrSessionData', 'label': visit_id,
                 'req_format': 'qa'}
        response = xnat_login.put(uri, query=query)
        if response.status_code not in (200, 201):
            raise ArcanaError(
                "Could not create session '{}' in subject '{}' in project '{}'"
                " response code {}"
                .format(visit_id, subject_id, self.inputs.project_id,
                        response))
        return xnat_login.classes.MrSessionData(uri=uri,
                                                xnat_session=xnat_login)


class XnatSink(XnatSinkMixin, ArchiveSink):

    input_spec = XnatSinkInputSpec


class XnatSubjectSink(XnatSinkMixin, ArchiveSubjectSink):

    input_spec = XnatSubjectSinkInputSpec


class XnatVisitSink(XnatSinkMixin, ArchiveVisitSink):

    input_spec = XnatVisitSinkInputSpec


class XnatProjectSink(XnatSinkMixin, ArchiveProjectSink):

    input_spec = XnatProjectSinkInputSpec


class XnatArchive(Archive):
    """
    An 'Archive' class for XNAT repositories

    Parameters
    ----------
    project_id : str
        The ID of the project on XNAT
    user : str
        Username with which to connect to XNAT with
    password : str
        Password to connect to XNAt with
    cache_dir : str (path)
        Path to local directory to cache XNAT data in
    server : str (URI)
        URI of XNAT server to connect to
    check_md5 : bool
        Whether to check the MD5 digest of cached files before using. This
        checks for updates on the server since the file was cached
    subject_ids : list(str) | None
        A list of subject IDs to filter the project search for. Will
        reduce the time taken to initialise the archive but will also
        limit the subjects that can be analysed
    visit_ids : list(str) | None
        A list of subject IDs to filter the project search for. Will
        reduce the time taken to initialise the archive but will also
        limit the subjects that can be analysed
    """

    type = 'xnat'
    Sink = XnatSink
    Source = XnatSource
    SubjectSink = XnatSubjectSink
    VisitSink = XnatVisitSink
    ProjectSink = XnatProjectSink

    SUMMARY_NAME = 'ALL'
    PROCESSED_SUFFIX = '_PROC'
    MD5_SUFFIX = '.md5.json'

    def __init__(self, server, project_id, user=None, password=None,
                 cache_dir=None, check_md5=True):
        self._project_id = project_id
        self._server = server
        self._user = user
        self._password = password
        if cache_dir is None:
            self._cache_dir = os.path.join(os.environ['HOME'], '.xnat')
        else:
            self._cache_dir = cache_dir
        try:
            # Attempt to make cache if it doesn't already exist
            os.makedirs(self._cache_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        self._check_md5 = check_md5

    def __repr__(self):
        return ("{}(server={}, project_id={}, cache_dir={})"
                .format(type(self).__name__,
                        self.server, self.project_id,
                        self._cache_dir))

    def __eq__(self, other):
        try:
            return (self.server == other.server and
                    self._cache_dir == other._cache_dir and
                    self.project_id == other.project_id)
        except AttributeError:
            return False  # For comparison with other types

    def source(self, *args, **kwargs):
        source = super(XnatArchive, self).source(*args, **kwargs)
        source.inputs.project_id = str(self.project_id)
        source.inputs.server = self._server
        if self._user is not None:
            source.inputs.user = self._user
        if self._password is not None:
            source.inputs.password = self._password
        source.inputs.cache_dir = self._cache_dir
        return source

    def sink(self, *args, **kwargs):
        sink = super(XnatArchive, self).sink(*args, **kwargs)
        sink.inputs.project_id = str(self.project_id)
        sink.inputs.server = self._server
        if self._user is not None:
            sink.inputs.user = self._user
        if self._password is not None:
            sink.inputs.password = self._password
        sink.inputs.cache_dir = self._cache_dir
        return sink

    def login(self, prev_login=None):
        """
        Parameters
        ----------
        prev_login : xnat.XNATSession
            An XNAT login that has been opened in the code that calls
            the method that calls login. It is wrapped in a
            NoExitWrapper so the returned connection can be used
            in a "with" statement in the method.
        """
        if prev_login is not None:
            return NoContextWrapper(prev_login)
        sess_kwargs = {}
        if self._user is not None:
            sess_kwargs['user'] = self._user
        if self._password is not None:
            sess_kwargs['password'] = self._password
        return xnat.connect(server=self._server, **sess_kwargs)

    def cache(self, dataset, prev_login=None):
        """
        Caches a single dataset (if the 'path' attribute is accessed
        and it has not been previously cached for example

        Parameters
        ----------
        dataset : Dataset
            The dataset to cache
        prev_login : xnat.XNATSession
            An XNATSession object to use for the connection. A new
            one is created if one isn't provided
        """
        if dataset.archive is not self:
            raise ArcanaError(
                "{} is not from {}".format(dataset, self))
        assert dataset.uri is not None
        with self.login(prev_login=prev_login) as xnat_login:
            sess_id, scan_id = re.match(
                r'/data/experiments/(\w+)/scans/(.*)',
                dataset.uri).groups()
            xsession = xnat_login.experiments[sess_id]
            xdataset = xsession.scans[scan_id]
            xresource = XnatSource.get_resource(xdataset, dataset)
            cache_path = self.cache_path(dataset)
            XnatSource.download_dataset(
                tempfile.mkdtemp(), xresource, xdataset, dataset,
                xsession.label, cache_path)
        return cache_path

    def cache_path(self, dataset):
        subj_dir, sess_dir = self.get_labels(
            dataset.frequency, self.project_id,
            dataset.subject_id, dataset.visit_id)
        return os.path.join(self._cache_dir, self.project_id,
                            subj_dir, sess_dir, dataset.fname())

    def all_session_ids(self, project_id):
        """
        Parameters
        ----------
        project_id : int
            The project id to return the sessions for
        repo_id : int
            The id of the repository (2 for monash daris)
        visit_ids: int|List[int]|None
            Id or ids of sessions of which to return sessions for. If None all
            are returned
        """
        sess_kwargs = {}
        if self._user is not None:
            sess_kwargs['user'] = self._user
        if self._password is not None:
            sess_kwargs['password'] = self._password
        with self.login() as xnat_login:
            return [
                s.label for s in xnat_login.projects[
                    project_id].experiments.itervalues()]

    def get_tree(self, subject_ids=None, visit_ids=None):
        """
        Return the tree of subject and sessions information within a
        project in the XNAT archive

        Parameters
        ----------
        subject_ids : list(str)
            List of subject IDs with which to filter the tree with. If
            None all are returned
        visit_ids : list(str)
            List of visit IDs with which to filter the tree with. If
            None all are returned

        Returns
        -------
        project : arcana.archive.Project
            A hierarchical tree of subject, session and dataset
            information for the archive
        """
        # Convert subject ids to strings if they are integers
        if subject_ids is not None:
            subject_ids = [
                ('{:03d}'.format(s)
                 if isinstance(s, int) else s) for s in subject_ids]
        # Add derived visit IDs to list of visit ids to filter
        if visit_ids is not None:
            visit_ids = visit_ids + [i + self.PROCESSED_SUFFIX
                                     for i in visit_ids]
        subjects = []
        sessions = defaultdict(list)
        with self.login() as xnat_login:
            xproject = xnat_login.projects[self.project_id]
            visit_sessions = defaultdict(list)
            # Create list of subjects
            for xsubject in xproject.subjects.itervalues():
                # This assumes that the subject ID is prepended with
                # the project ID
                subj_id = xsubject.label[(len(self.project_id) + 1):]
                if subj_id == XnatArchive.SUMMARY_NAME:
                    continue
                if (subject_ids is not None and
                        subj_id not in subject_ids):
                    continue
                logger.debug("Getting info for subject '{}'"
                             .format(subj_id))
                sessions = {}
                proc_sessions = []
                # Get per_session datasets
                for xsession in xsubject.experiments.itervalues():
                    visit_id = '_'.join(xsession.label.split('_')[2:])
                    if visit_id == XnatArchive.SUMMARY_NAME:
                        continue
                    if not (visit_ids is None or visit_id in visit_ids):
                        continue
                    derived = xsession.label.endswith(
                        self.PROCESSED_SUFFIX)
                    session = Session(subj_id, visit_id,
                                      datasets=self._get_datasets(
                                          xsession, 'per_session',
                                          subject_id=subj_id,
                                          visit_id=visit_id,
                                          derived=derived),
                                      fields=self._get_fields(
                                          xsession, 'per_session',
                                          subject_id=subj_id,
                                          visit_id=visit_id,
                                          derived=derived),
                                      derived=None)
                    if derived:
                        proc_sessions.append(session)
                    else:
                        sessions[visit_id] = session
                        visit_sessions[visit_id].append(session)
                for proc_session in proc_sessions:
                    visit_id = proc_session.visit_id[:-len(
                        self.PROCESSED_SUFFIX)]
                    try:
                        sessions[visit_id].derived = proc_session
                    except KeyError:
                        raise ArcanaError(
                            "No matching acquired session for derived "
                            "session '{}_{}_{}'".format(
                                self.project_id,
                                proc_session.subject_id,
                                proc_session.visit_id))
                # Get per_subject datasets
                subj_summary_name = self.get_labels(
                    'per_subject', self.project_id, subj_id)[1]
                try:
                    xsubj_summary = xsubject.experiments[
                        subj_summary_name]
                except KeyError:
                    subj_datasets = []
                    subj_fields = []
                else:
                    subj_datasets = self._get_datasets(
                        xsubj_summary, 'per_subject',
                        subject_id=subj_id)
                    subj_fields = self._get_fields(
                        xsubj_summary, 'per_subject',
                        subject_id=subj_id)
                subjects.append(Subject(subj_id,
                                        sorted(sessions.values()),
                                        datasets=subj_datasets,
                                        fields=subj_fields))
            # Create list of visits
            visits = []
            for visit_id, v_sessions in visit_sessions.iteritems():
                (_, visit_summary_sess_name) = self.get_labels(
                    'per_visit', self.project_id, visit_id=visit_id)
                # Get 'per_visit' datasets
                try:
                    xvisit_summary = xproject.experiments[
                        visit_summary_sess_name]
                except KeyError:
                    visit_datasets = []
                    visit_fields = {}
                else:
                    visit_datasets = self._get_datasets(
                        xvisit_summary, 'per_visit', visit_id=visit_id)
                    visit_fields = self._get_fields(
                        xvisit_summary, 'per_visit', visit_id=visit_id)
                visits.append(Visit(visit_id, sorted(v_sessions),
                                    datasets=visit_datasets,
                                    fields=visit_fields))
            # Get 'per_project' datasets
            (proj_summary_subj_name,
             proj_summary_sess_name) = self.get_labels('per_project',
                                                       self.project_id)
            try:
                xproj_summary = xproject.subjects[
                    proj_summary_subj_name].experiments[proj_summary_sess_name]
            except KeyError:
                proj_datasets = []
                proj_fields = []
            else:
                proj_datasets = self._get_datasets(xproj_summary,
                                                   'per_project')
                proj_fields = self._get_fields(xproj_summary,
                                               'per_project')
            if not subjects:
                raise ArcanaError(
                    "Did not find any subjects matching the IDs '{}' in "
                    "project '{}' (found '{}')"
                    .format(
                        ("', '".join(subject_ids)
                         if subject_ids is not None else ''),
                        self.project_id,
                        "', '".join(s.label[(len(self.project_id) + 1):]
                                    for s in xproject.subjects.values())))
            if not sessions:
                raise ArcanaError(
                    "Did not find any sessions matching the visit IDs "
                    "'{}' (in subjects '{}') for project '{}'"
                    .format(
                        ("', '".join(visit_ids)
                         if visit_ids is not None else ''),
                        "', '".join(
                            s.label.split('_')[1]
                            for s in xproject.experiments.values()),
                        self.project_id))
        return Project(sorted(subjects), sorted(visits),
                       datasets=proj_datasets, fields=proj_fields)

    def _get_datasets(self, xsession, freq, subject_id=None,
                      visit_id=None, derived=False):
        """
        Returns a list of datasets within an XNAT session

        Parameters
        ----------
        xsession : xnat.classes.MrSessionData
            The XNAT session to extract the datasets from
        freq : str
            The frequency of the returned datasets (either 'per_session',
            'per_subject', 'per_visit', or 'per_project')
        derived : bool
            Whether the session is derived or not

        Returns
        -------
        datasets : list(arcana.dataset.Dataset)
            List of datasets within an XNAT session
        """
        datasets = []
        for xdataset in xsession.scans.itervalues():
            data_format = DataFormat.by_name(
                guess_data_format(xdataset))
            datasets.append(Dataset(
                xdataset.type, format=data_format, derived=derived,  # @ReservedAssignment @IgnorePep8
                frequency=freq, path=None, id=xdataset.id,
                uri=xdataset.uri, subject_id=subject_id,
                visit_id=visit_id, archive=self))
        return sorted(datasets)

    def _get_fields(self, xsession, freq, subject_id=None,
                    visit_id=None, derived=False):
        """
        Returns a list of fields within an XNAT session

        Parameters
        ----------
        xsession : xnat.classes.MrSessionData
            The XNAT session to extract the fields from
        freq : str
            The frequency of the returned fields (either 'per_session',
            'per_subject', 'per_visit', or 'per_project')

        Returns
        -------
        fields : list(arcana.dataset.Field)
            List of fields within an XNAT session
        """
        fields = []
        for name, value in xsession.fields.items():
            fields.append(Field(
                name=name, value=value, derived=derived,
                frequency=freq, subject_id=subject_id,
                visit_id=visit_id, archive=self))
        return sorted(fields)

    def dicom_header(self, dataset, prev_login=None):
        with self.login(prev_login) as xnat_login:
            response = xnat_login.get(
                '/REST/services/dicomdump?src=/archive/projects/{}'
                '{}&format=json'
                .format(self.project_id, dataset.uri[len('/data'):]))
        def convert(val, code):  # @IgnorePep8
            if code == 'TM':
                val = float(val)
            elif code == 'CS':
                val = val.split('\\')
            return val
        hdr = {tag_parse_re.match(t['tag1']).groups():
               convert(t['value'], t['vr'])
               for t in response.json()['ResultSet']['Result']
               if (tag_parse_re.match(t['tag1']) and
                   t['vr'] in RELEVANT_DICOM_TAG_TYPES)}
        return hdr

    @property
    def project_id(self):
        return self._project_id

    @property
    def server(self):
        return self._server

    @property
    def cache_dir(self):
        return self._cache_dir

    @classmethod
    def get_labels(cls, frequency, project_id, subject_id=None,
                   visit_id=None):
        """
        Returns the labels for the XNAT subject and sessions given
        the frequency and provided IDs.
        """
        if frequency == 'per_session':
            assert visit_id is not None
            assert subject_id is not None
            subj_label = '{}_{}'.format(project_id, subject_id)
            sess_label = '{}_{}_{}'.format(project_id, subject_id,
                                           visit_id)
        elif frequency == 'per_subject':
            assert subject_id is not None
            subj_label = '{}_{}'.format(project_id, subject_id)
            sess_label = '{}_{}_{}'.format(project_id, subject_id,
                                           cls.SUMMARY_NAME)
        elif frequency == 'per_visit':
            assert visit_id is not None
            subj_label = '{}_{}'.format(project_id, cls.SUMMARY_NAME)
            sess_label = '{}_{}_{}'.format(project_id, cls.SUMMARY_NAME,
                                           visit_id)
        elif frequency == 'per_project':
            subj_label = '{}_{}'.format(project_id, cls.SUMMARY_NAME)
            sess_label = '{}_{}_{}'.format(project_id, cls.SUMMARY_NAME,
                                           cls.SUMMARY_NAME)
        else:
            raise ArcanaError(
                "Unrecognised frequency '{}'".format(frequency))
        return (subj_label, sess_label)


def guess_data_format(dataset):
    dataset_formats = [r for r in dataset.resources.itervalues()
                       if r.label.lower() in DataFormat.by_names]
    if len(dataset_formats) > 1:
        raise ArcanaError(
            "Multiple valid resources '{}' for '{}' dataset, please pass "
            "'data_format' to 'download_dataset' method to speficy resource to"
            "download".format("', '".join(dataset_formats), dataset.type))
    elif not dataset_formats:
        raise ArcanaError(
            "No recognised data formats for '{}' dataset (available resources "
            "are '{}')".format(
                dataset.type, "', '".join(
                    r.label for r in dataset.resources.itervalues())))
    return dataset_formats[0].label



from abc import ABCMeta, abstractmethod
from copy import copy
from arcana.node import Node
from arcana.interfaces.utils import (
    ZipDir, UnzipDir, TarGzDir, UnTarGzDir)
from arcana.exception import (
    ArcanaRequirementVersionException,
    ArcanaModulesNotInstalledException,
    ArcanaUsageError, ArcanaDataFormatClashError,
    ArcanaNoConverterError,
    ArcanaConverterNotAvailableError,
    ArcanaDataFormatNotRegisteredError)
from nipype.interfaces.utility import IdentityInterface
from arcana.requirement import Requirement
import logging


logger = logging.getLogger('Arcana')


class DataFormat(object):
    """
    Defines a format for a dataset (e.g. DICOM, NIfTI, Matlab file)

    Parameters
    ----------
    name : str
        A name for the data format
    extension : str
        The extension of the format
    desc : str
        A description of what the format is and ideally a link to its
        documentation
    directory : bool
        Whether the format is a directory or a file
    within_dir_exts : List[str]
        A list of extensions that are found within the top level of
        the directory (for directory formats). Used to identify
        formats from paths.
    converters : Dict[str, Converter]
        A dictionary mapping names of alternative data formats
        to Converter objects that can convert from the alternative
        format to this format.
    """

    # To hold registered data formats
    by_names = {}
    by_exts = {}
    by_within_exts = {}

    def __init__(self, name, extension=None, desc='',
                 directory=False, within_dir_exts=None,
                 converters=None):
        if not name.islower():
            raise ArcanaUsageError(
                "All data format names must be lower case ('{}')"
                .format(name))
        if extension is None and not directory:
            raise ArcanaUsageError(
                "Extension for '{}' format can only be None if it is a "
                "directory".format(name))
        self._name = name
        self._extension = extension
        self._desc = desc
        self._directory = directory
        if within_dir_exts is not None:
            if not directory:
                raise ArcanaUsageError(
                    "'within_dir_exts' keyword arg is only valid "
                    "for directory data formats, not '{}'".format(name))
            within_dir_exts = frozenset(within_dir_exts)
        self._within_dir_exts = within_dir_exts
        self._converters = converters if converters is not None else {}

    def __eq__(self, other):
        try:
            return (
                self._name == other._name and
                self._extension == other._extension and
                self._desc == other._desc and
                self._directory == other._directory and
                self._within_dir_exts ==
                other._within_dir_exts and
                self._converters == other._converters)
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return ("DataFormat(name='{}', extension='{}', directory={}{})"
                .format(self.name, self.extension, self.directory,
                        ('within_dir_extension={}'.format(
                            self.within_dir_exts)
                         if self.directory else '')))

    def __str__(self):
        return self.name

    @property
    def name(self):
        return self._name

    @property
    def extension(self):
        return self._extension

    @property
    def ext(self):
        return self._extension

    @property
    def ext_str(self):
        return self.extension if self.extension is not None else ''

    @property
    def desc(self):
        return self._desc

    @property
    def directory(self):
        return self._directory

    @property
    def within_dir_exts(self):
        return self._within_dir_exts

    @property
    def xnat_resource_name(self):
        return self.name.upper()

    def converter_from(self, data_format):
        if data_format == self:
            return IdentityConverter(data_format, self)
        try:
            converter_cls = self._converters[data_format.name]
        except KeyError:
            raise ArcanaNoConverterError(
                "There is no converter to convert {} to {}"
                .format(self, data_format))
        return converter_cls(data_format, self)

    @classmethod
    def register(cls, data_format):
        """
        Registers a data format so they can be recognised by extension
        and by resource type on XNAT

        Parameters
        ----------
        data_format : DataFormat
            The data format to register
        """
        try:
            saved_format = cls.by_names[data_format.name]
            if saved_format != data_format:
                raise ArcanaDataFormatClashError(
                    "Cannot register {} due to name clash with previously "
                    "registered {}".format(data_format, saved_format))
        except KeyError:
            if data_format.directory and data_format.extension is None:
                if data_format.within_dir_exts in cls.by_within_exts:
                    raise ArcanaDataFormatClashError(
                        "Cannot register {} due to within-directory "
                        "extension clash with previously registered {}"
                        .format(data_format,
                                cls.by_within_exts[
                                    data_format.within_dir_exts]))
            else:
                if data_format.extension in cls.by_exts:
                    raise ArcanaDataFormatClashError(
                        "Cannot register {} due to extension clash with "
                        "previously registered {}".format(
                            data_format,
                            cls.by_exts[data_format.ext]))
            cls.by_names[data_format.name] = data_format
            if data_format.ext is not None:
                cls.by_exts[data_format.ext] = data_format
            if data_format.within_dir_exts is not None:
                cls.by_within_exts[
                    data_format.within_dir_exts] = data_format

    @classmethod
    def by_name(cls, name):
        try:
            return cls.by_names[name.lower()]
        except KeyError:
            raise ArcanaDataFormatNotRegisteredError(
                "No data format named '{}' has been registered"
                .format(name,
                        ', '.format(repr(f)
                                    for f in cls.by_names.values())))

    @classmethod
    def by_ext(cls, ext):
        try:
            return cls.by_exts[ext]
        except KeyError:
            raise ArcanaDataFormatNotRegisteredError(
                "No data format with extension '{}' has been registered"
                .format(
                    ext, ', '.format(repr(f)
                                     for f in cls.by_exts.values())))

    @classmethod
    def by_within_dir_exts(cls, within_exts):
        try:
            return cls.by_within_exts[within_exts]
        except KeyError:
            raise ArcanaDataFormatNotRegisteredError(
                "No data format with within-directory extension '{}' "
                "has been registered ({})".format(
                    within_exts,
                    ', '.format(repr(f)
                                for f in cls.by_within_exts.values())))


class Converter(object):
    """
    Base class for all Arcana data format converters

    Parameters
    ----------
    input_format : DataFormat
        The input format to convert from
    output_format : DataFormat
        The output format to convert to
    """

    __metaclass__ = ABCMeta

    def __init__(self, input_format, output_format):
        self._input_format = input_format
        self._output_format = output_format
        try:
            available_modules = Node.available_modules()
            for possible_reqs in self.requirements:
                Requirement.best_requirement(possible_reqs,
                                             available_modules)
        except ArcanaRequirementVersionException:
            raise ArcanaConverterNotAvailableError(
                "Module(s) required for converter {} ({}) are not "
                "available".format(
                    self,
                    ', '.join(r.name for r in self.requirements)))
        except ArcanaModulesNotInstalledException:
            pass

    @property
    def input_format(self):
        return self._input_format

    @property
    def output_format(self):
        return self._output_format

    @abstractmethod
    def get_node(self, name):
        """
        Returns a NiPype node that converts a dataset from the input
        format to the output format

        Parameters
        ----------
        name : str
            Name for the node
        """

    def __repr__(self):
        return "{}(input_format={}, output_format={})".format(
            type(self).__name__, self.input_format, self.output_format)


class IdentityConverter(Converter):

    requirements = []

    def get_node(self, name):
        return Node(IdentityInterface(['i']), name=name), 'i', 'i'


class UnzipConverter(Converter):

    requirements = []

    def get_node(self, name):
        convert_node = Node(UnzipDir(), name=name, memory=12000)
        return convert_node, 'zipped', 'unzipped'


class ZipConverter(Converter):

    requirements = []

    def get_node(self, name):
        convert_node = Node(ZipDir(), name=name, memory=12000)
        return convert_node, 'dirname', 'zipped'


class TarGzConverter(Converter):

    requirements = []

    def get_node(self, name):
        convert_node = Node(TarGzDir(), name=name, memory=12000)
        return convert_node, 'dirname', 'zipped'


class UnTarGzConverter(Converter):

    requirements = []

    def get_node(self, name):
        convert_node = Node(UnTarGzDir(), name=name, memory=12000)
        return convert_node, 'gzipped', 'gunzipped'


# General formats
directory_format = DataFormat(name='directory', extension=None,
                              directory=True,
                              converters={'zip': UnzipConverter,
                                          'targz': UnTarGzConverter})
text_format = DataFormat(name='text', extension='.txt')


# Compressed formats
zip_format = DataFormat(name='zip', extension='.zip',
                        converters={'directory': ZipConverter})
targz_format = DataFormat(name='targz', extension='.tar.gz',
                          converters={'direcctory': TarGzConverter})

# Register all data formats in module
for data_format in copy(globals()).itervalues():
    if isinstance(data_format, DataFormat):
        DataFormat.register(data_format)



import re
from itertools import izip_longest
from arcana.exception import (
    ArcanaError, ArcanaRequirementVersionException)
import logging

logger = logging.getLogger('Arcana')


def split_version(version_str):
    logger.debug("splitting version string '{}'"
                 .format(version_str))
    try:
        sanitized_ver_str = re.match(r'[^\d]*(\d+(?:\.\d+)*)[^\d]*',
                                     version_str).group(1)
        return tuple(
            int(p) for p in sanitized_ver_str.split('.'))
    except (ValueError, AttributeError) as e:
        raise ArcanaRequirementVersionException(
            "Could not parse version string '{}': {}".format(
                version_str, e))


def date_split(version_str):
    try:
        return tuple(int(p) for p in version_str.split('-'))
    except ValueError as e:
        raise ArcanaRequirementVersionException(str(e))


def matlab_version_split(version_str):
    match = re.match(r'(?:r|R)?(\d+)(\w)', version_str)
    if match is None:
        raise ArcanaRequirementVersionException(
            "Do not understand Matlab version '{}'".format(version_str))
    return int(match.group(1)), match.group(2).lower()


class Requirement(object):
    """
    Defines a software package that is required by a processing Node, which is
    typically wrapped up in an environment module (see
    http://modules.sourceforge.net)

    Parameters
    ----------
    name : str
        Name of the package
    min_version : tuple(int|str)
        The minimum version required by the node
    max_version : tuple(int|str) | None
        The maximum version that is compatible with the Node
    """

    def __init__(self, name, min_version, max_version=None,
                 version_split=split_version, citations=None):
        self._name = name.lower()
        self._min_ver = tuple(min_version)
        if max_version is not None:
            self._max_ver = tuple(max_version)
            if not self.later_or_equal_version(self._max_ver, self._min_ver):
                raise ArcanaError(
                    "Supplied max version ({}) is not greater than min "
                    " version ({})".format(self._min_ver, self._max_ver))
        else:
            self._max_ver = None
        self._version_split = version_split
        self._citations = citations if citations is not None else []

    @property
    def name(self):
        return self._name

    def __repr__(self):
        return "{} (v >= {}{})".format(
            self.name, self.min_version,
            (", v <= {}".format(self.max_version)
             if self.max_version is not None else ''))

    @property
    def min_version(self):
        return self._min_ver

    @property
    def citations(self):
        return iter(self._citations)

    @property
    def max_version(self):
        return self._max_ver

    def split_version(self, version_str):
        return tuple(self._version_split(version_str))

    def best_version(self, available_versions):
        """
        Picks the latest acceptible version from the versions available

        Parameters
        ----------
        available_versions : list(str)
            List of possible versions
        """
        best = None
        for ver in available_versions:
            try:
                v_parts = self.split_version(ver)
            except ArcanaRequirementVersionException:
                continue  # Incompatible version
            if (self.later_or_equal_version(v_parts, self._min_ver) and
                (self._max_ver is None or
                 self.later_or_equal_version(self._max_ver, v_parts))):
                if best is None or self.later_or_equal_version(v_parts,
                                                               best[1]):
                    best = ver, v_parts
        if best is None:
            msg = ("Could not find version of '{}' matching requirements "
                   "> ({})"
                   .format(self.name,
                           ', '.join(str(v) for v in self._min_ver)))
            if self._max_ver is not None:
                msg += " and < ({})".format(
                    ', '.join(str(v) for v in self._max_ver))
            msg += " from available versions '{}'".format(
                "', '".join(available_versions))
            raise ArcanaRequirementVersionException(msg)
        return best[0]

    def valid_version(self, version):
        return (self.later_or_equal_version(version, self.min_version) and
                (self.max_version is None or
                 self.later_or_equal_version(self.max_version, version)))

    @classmethod
    def later_or_equal_version(cls, version, reference):
        for v_part, r_part in izip_longest(version, reference, fillvalue=0):
            if type(v_part) != type(r_part):
                raise ArcanaError(
                    "Type of version part {} (of '{}'), {}, does not match "
                    "type of reference part {}, {}".format(
                        v_part, version, type(v_part), r_part, type(r_part)))
            if v_part > r_part:
                return True
            elif v_part < r_part:
                return False
        return True

    @classmethod
    def best_requirement(cls, possible_requirements, available_modules,
                         preloaded_modules=None):
        if preloaded_modules is None:
            preloaded_modules = {}
        # If possible reqs is a singleton, wrap it in a list for
        # iterating
        if isinstance(possible_requirements, Requirement):
            possible_requirements = [possible_requirements]
        # Loop through all options for a given requirement and see
        # if at least one can be satisfied.
        logger.debug(
            "Searching for one of {}".format(
                ', '.join(str(r) for r in possible_requirements)))
        ver_exceptions = []  # Will hold all version error messages
        for req in possible_requirements:
            try:
                version = preloaded_modules[req.name]
                logger.debug("Found preloaded version {} of module '{}'"
                             .format(version, req.name))
                if req.valid_version(req.split_version(version)):
                    return req.name, version
                else:
                    raise ArcanaError(
                        "Incompatible module version already loaded {}/{},"
                        " (valid {}->{}) please unload before running "
                        "pipeline"
                        .format(
                            req.name, version, req.min_version,
                            (req.max_version if req.max_version is not None
                             else '')))
            except KeyError:
                try:
                    best_version = req.best_version(
                        available_modules[req.name])
                    logger.debug("Found best version '{}' of module '{}' for"
                                 " requirement {}".format(best_version,
                                                          req.name, req))
                    return req.name, best_version
                except ArcanaRequirementVersionException as e:
                    ver_exceptions.append(e)
        # If no options can be satisfied, otherwise raise exception with
        # combined messages from all options.
        raise ArcanaRequirementVersionException(
            ' and '.join(str(e) for e in ver_exceptions))



import os.path
from arcana.exception import ArcanaError
import re


PATH_SUFFIX = '_path'
FIELD_SUFFIX = '_field'

package_dir = os.path.join(os.path.dirname(__file__), '..')


def dir_modtime(dpath):
    """
    Returns the latest modification time of all files/subdirectories in a
    directory
    """
    return max(os.path.getmtime(d) for d, _, _ in os.walk(dpath))


double_exts = ('.tar.gz', '.nii.gz')


def split_extension(path):
    """
    A extension splitter that checks for compound extensions such as
    'file.nii.gz'

    Parameters
    ----------
    filename : str
        A filename to split into base and extension

    Returns
    -------
    base : str
        The base part of the string, i.e. 'file' of 'file.nii.gz'
    ext : str
        The extension part of the string, i.e. 'nii.gz' of 'file.nii.gz'
    """
    for double_ext in double_exts:
        if path.endswith(double_ext):
            return path[:-len(double_ext)], double_ext
    dirname = os.path.dirname(path)
    filename = os.path.basename(path)
    parts = filename.split('.')
    if len(parts) == 1:
        base = filename
        ext = None
    else:
        ext = '.' + parts[-1]
        base = '.'.join(parts[:-1])
    return os.path.join(dirname, base), ext


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class NoContextWrapper(object):
    """
    Wraps an object, passing all calls through to the wrapped object
    except the __enter__ and __exit__ method, which do nothing. Used
    in cases where you want to use a file|connection handle within a
    "with" statement, except when it passed to the method from the
    calling code (presumably nested in another "with" statement).
    """

    def __init__(self, to_wrap):
        self._to_wrap = to_wrap

    def __getattr__(self, name):
        return getattr(self._to_wrap, name)

    def __enter__(self, *args, **kwargs):  # @UnusedVariable
        return self

    def __exit__(self, *args, **kwargs):
        pass



from copy import copy
from nipype.pipeline import engine as pe


class BaseRunner(object):
    """
    A thin wrapper around the NiPype LinearPlugin used to connect
    runs pipelines on the local workstation

    Parameters
    ----------
    work_dir : str
        A directory in which to run the nipype workflows
    max_process_time : float
        The maximum time allowed for the process
    reprocess: True|False|'all'
        A flag which determines whether to rerun the processing for this
        step. If set to 'all' then pre-requisite pipelines will also be
        reprocessed.
    """

    default_plugin_args = {}

    def __init__(self, work_dir, max_process_time=None,
                 reprocess=False, **kwargs):
        self._work_dir = work_dir
        self._max_process_time = max_process_time
        self._reprocess = reprocess
        self._plugin_args = copy(self.default_plugin_args)
        self._plugin_args.update(kwargs)
        self._init_plugin()

    def _init_plugin(self):
        self._plugin = self.nipype_plugin_cls(**self._plugin_args)

    def run(self, pipeline, **kwargs):
        workflow = pe.Workflow(name=pipeline.name,
                               base_dir=self.work_dir)
        pipeline.connect_to_archive(workflow, reprocess=self._reprocess,
                                    **kwargs)
        # Reset the cached tree of datasets in the archive as it will
        # change after the pipeline has run.
        pipeline.study.reset_tree()
        return workflow.run(plugin=self._plugin)

    def __repr__(self):
        return "{}(work_dir={})".format(
            type(self).__name__, self._work_dir)

    def __eq__(self, other):
        try:
            return (self._work_dir == other._work_dir and
                    (self._max_process_time ==
                     other._max_process_time) and
                    self._plugin_args == other._plugin_args)
        except AttributeError:
            return False

    @property
    def work_dir(self):
        return self._work_dir

    def __getstate__(self):
        dct = copy(self.__dict__)
        # Delete the NiPype plugin as it can be regenerated
        del dct['_plugin']
        return dct

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._init_plugin()



from .base import BaseRunner
from nipype.pipeline.plugins import MultiProcPlugin


class MultiProcRunner(BaseRunner):
    """
    A thin wrapper around the NiPype MultiProcPlugin used to
    run pipelines on the local workstation on muliple processes

    Parameters
    ----------
    work_dir : str
        A directory in which to run the nipype workflows
    num_processes : int
        The number of processes to use
    max_process_time : float
        The maximum time allowed for the process
    reprocess: True|False|'all'
        A flag which determines whether to rerun the processing for this
        step. If set to 'all' then pre-requisite pipelines will also be
        reprocessed.
    """

    nipype_plugin_cls = MultiProcPlugin

    def __init__(self, work_dir, num_processes=None, plugin_args=None,
                 **kwargs):
        if plugin_args is None:
            plugin_args = {}
        if num_processes is not None:
            plugin_args['n_procs'] = num_processes
        super(MultiProcRunner, self).__init__(
            work_dir, plugin_args=plugin_args, **kwargs)

    @property
    def num_processes(self):
        return self._plugin.processes



import os.path as op
import os
import subprocess as sp
import operator
import shutil
from unittest import TestCase
from functools import reduce
import errno
import sys
import json
import filecmp
from copy import deepcopy
import logging
import arcana
from arcana.dataset import Dataset
from arcana.utils import classproperty
from arcana.archive.local import (
    LocalArchive, SUMMARY_NAME)
from arcana.runner import LinearRunner
from arcana.exception import ArcanaError
from arcana.node import ArcanaNodeMixin
from arcana.exception import (
    ArcanaModulesNotInstalledException)
from arcana.archive.local import (
    SUMMARY_NAME as LOCAL_SUMMARY_NAME, FIELDS_FNAME)
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterface, isdefined)

logger = logging.getLogger('Arcana')
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

logging.getLogger("urllib3").setLevel(logging.WARNING)


class BaseTestCase(TestCase):

    SUBJECT = 'SUBJECT'
    VISIT = 'VISIT'

    # Whether to copy reference datasets from reference directory
    INPUTS_FROM_REF_DIR = False

    # The path to the test directory, which should sit along side the
    # the package directory. Note this will not work when Arcana
    # is installed by a package manager.
    BASE_TEST_DIR = op.abspath(op.join(
        op.dirname(arcana.__file__), '..', 'test'))

    @classproperty
    @classmethod
    def test_data_dir(cls):
        return op.join(cls.BASE_TEST_DIR, 'data')

    @classproperty
    @classmethod
    def unittest_root(cls):
        return op.join(cls.BASE_TEST_DIR, 'unittests')

    @classproperty
    @classmethod
    def archive_path(cls):
        return op.join(cls.test_data_dir, 'archive')

    @classproperty
    @classmethod
    def work_path(cls):
        return op.join(cls.test_data_dir, 'work')

    @classproperty
    @classmethod
    def cache_path(cls):
        return op.join(cls.test_data_dir, 'cache')

    def setUp(self):
        self.reset_dirs()
        if self.INPUTS_FROM_REF_DIR:
            datasets = {}
            for fname in os.listdir(self.ref_dir):
                if fname.startswith('.'):
                    continue
                dataset = Dataset.from_path(op.join(self.ref_dir,
                                                    fname))
                datasets[dataset.name] = dataset
        else:
            datasets = getattr(self, 'INPUT_DATASETS', None)
        self.add_session(datasets=datasets,
                         fields=getattr(self, 'INPUT_FIELDS', None))

    def add_session(self, datasets=None, fields=None, project_dir=None,
                    subject=None, visit=None):
        if project_dir is None:
            project_dir = self.project_dir
        if datasets is None:
            datasets = {}
        if subject is None:
            subject = self.SUBJECT
        if visit is None:
            visit = self.VISIT
        session_dir = op.join(project_dir, subject, visit)
        os.makedirs(session_dir)
        for name, dataset in datasets.items():
            if isinstance(dataset, Dataset):
                dst_path = op.join(session_dir,
                                   name + dataset.format.ext_str)
                if dataset.format.directory:
                    shutil.copytree(dataset.path, dst_path)
                else:
                    shutil.copy(dataset.path, dst_path)
            elif isinstance(dataset, basestring):
                # Write string as text file
                with open(op.join(session_dir,
                                  name + '.txt'), 'w') as f:
                    f.write(dataset)
            else:
                raise ArcanaError(
                    "Unrecognised dataset ({}) in {} test setup. Can "
                    "be either a Dataset or basestring object"
                    .format(dataset, self))
        if fields is not None:
            with open(op.join(session_dir,
                                   FIELDS_FNAME), 'w') as f:
                json.dump(fields, f)

    def delete_project(self, project_dir):
        # Clean out any existing archive files
        shutil.rmtree(project_dir, ignore_errors=True)

    def reset_dirs(self):
        shutil.rmtree(self.project_dir, ignore_errors=True)
        shutil.rmtree(self.work_dir, ignore_errors=True)
        self.create_dirs()

    def create_dirs(self):
        for d in (self.project_dir, self.work_dir):
            if not op.exists(d):
                os.makedirs(d)

    @property
    def archive_tree(self):
        return self.archive.get_tree()

    @property
    def xnat_session_name(self):
        return '{}_{}'.format(self.XNAT_TEST_PROJECT, self.name)

    @property
    def session_dir(self):
        return self.get_session_dir(self.SUBJECT, self.VISIT)

    @property
    def session(self):
        return self.archive_tree.subject(
            self.SUBJECT).session(self.VISIT)

    @property
    def archive(self):
        return self.local_archive

    @property
    def local_archive(self):
        try:
            return self._local_archive
        except AttributeError:
            self._local_archive = LocalArchive(self.project_dir)
            return self._local_archive

    @property
    def runner(self):
        return LinearRunner(self.work_dir)

    @property
    def project_dir(self):
        return op.join(self.archive_path, self.name)

    @property
    def work_dir(self):
        return op.join(self.work_path, self.name)

    @property
    def cache_dir(self):
        return op.join(self.cache_path, self.name)

    @property
    def ref_dir(self):
        return op.join(self.test_data_dir, 'reference', self.name)

    @property
    def name(self):
        return self._get_name(type(self))

    @property
    def project_id(self):
        return self.name  # To allow override in deriving classes

    def _get_name(self, cls):
        """
        Get unique name for test class from module path and its class name to
        be used for storing test data on XNAT and creating unique work/project
        dirs
        """
        module_path = op.abspath(sys.modules[cls.__module__].__file__)
        rel_module_path = module_path[(len(self.unittest_root) + 1):]
        path_parts = rel_module_path.split(op.sep)
        module_name = (''.join(path_parts[:-1]) +
                       op.splitext(path_parts[-1])[0][5:]).upper()
        test_class_name = cls.__name__[4:].upper()
        return module_name + '_' + test_class_name

    def create_study(self, study_cls, name, inputs, archive=None,
                     runner=None, **kwargs):
        """
        Creates a study using default archive and runners.

        Parameters
        ----------
        study_cls : Study
            The class to initialise
        name : str
            Name of the study
        inputs : List[BaseSpec]
            List of inputs to the study
        archive : BaseArchive | None
            The archive to use (a default local archive is used if one
            isn't provided
        runner : Runner | None
            The runner to use (a default LinearRunner is used if one
            isn't provided
        """
        if archive is None:
            archive = self.archive
        if runner is None:
            runner = self.runner
        return study_cls(
            name=name,
            archive=archive,
            runner=runner,
            inputs=inputs,
            **kwargs)

    def assertDatasetCreated(self, dataset_name, study_name, subject=None,
                             visit=None, frequency='per_session'):
        output_dir = self.get_session_dir(subject, visit, frequency)
        out_path = self.output_file_path(
            dataset_name, study_name, subject, visit, frequency)
        self.assertTrue(
            op.exists(out_path),
            ("Dataset '{}' (expected at '{}') was not created by unittest"
             " ('{}' found in '{}' instead)".format(
                 dataset_name, out_path, "', '".join(os.listdir(output_dir)),
                 output_dir)))

    def assertContentsEqual(self, dataset, reference, context=None):
        with open(dataset.path) as f:
            contents = f.read()
        msg = ("Contents of {} ({}) do not match reference ({})"
               .format(dataset, contents, reference))
        if context is not None:
            msg += 'for ' + context
        self.assertEqual(contents, str(reference), msg)

    def assertCreated(self, dataset):
        self.assertTrue(
            os.path.exists(dataset.path),
            "{} was not created".format(dataset))

    def assertField(self, name, ref_value, study_name, subject=None,
                    visit=None, frequency='per_session',
                    to_places=None):
        esc_name = study_name + '_' + name
        output_dir = self.get_session_dir(subject, visit, frequency)
        try:
            with open(op.join(output_dir, FIELDS_FNAME)) as f:
                fields = json.load(f)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise ArcanaError(
                    "No fields were created by pipeline in study '{}'"
                    .format(study_name))
        try:
            value = fields[esc_name]
        except KeyError:
            raise ArcanaError(
                "Field '{}' was not created by pipeline in study '{}'. "
                "Created fields were ('{}')"
                .format(esc_name, study_name, "', '".join(fields)))
        msg = ("Field value '{}' for study '{}', {}, does not match "
               "reference value ({})".format(name, study_name, value,
                                             ref_value))
        if to_places is not None:
            self.assertAlmostEqual(
                value, ref_value, to_places,
                '{} to {} decimal places'.format(msg, to_places))
        else:
            self.assertEqual(value, ref_value, msg)

    def assertDatasetsEqual(self, dataset1, dataset2, error_msg=None):
        msg = "{} does not match {}".format(dataset1, dataset2)
        if msg is not None:
            msg += ':\n' + error_msg
        self.assertTrue(filecmp.cmp(dataset1.path, dataset2.path,
                                    shallow=False), msg=msg)

    def assertStatEqual(self, stat, dataset_name, target, study_name,
                        subject=None, visit=None,
                        frequency='per_session'):
            try:
                ArcanaNodeMixin.load_module('mrtrix')
            except ArcanaModulesNotInstalledException:
                pass
            val = float(sp.check_output(
                'mrstats {} -output {}'.format(
                    self.output_file_path(
                        dataset_name, study_name,
                        subject=subject, visit=visit,
                        frequency=frequency),
                    stat),
                shell=True))
            self.assertEqual(
                val, target, (
                    "{} value of '{}' ({}) does not equal target ({}) "
                    "for subject {} visit {}"
                    .format(stat, dataset_name, val, target,
                            subject, visit)))

    def assertImagesAlmostMatch(self, out, ref, mean_threshold,
                                stdev_threshold, study_name):
        out_path = self.output_file_path(out, study_name)
        ref_path = self.ref_file_path(ref)
        # Should probably look into ITK fuzzy matching methods
        cmd = ("mrcalc -quiet {a} {b} -subtract - | mrstats - | "
               "grep -v channel | awk '{{print $4 \" \" $6}}'"
               .format(a=out_path, b=ref_path))
        out = sp.check_output(cmd, shell=True)
        mean, stdev = (float(x) for x in out.split())
        self.assert_(
            abs(mean) < mean_threshold and stdev < stdev_threshold,
            ("Mean ({mean}) or standard deviation ({stdev}) of difference "
             "between images {a} and {b} differ more than threshold(s) "
             "({thresh_mean} and {thresh_stdev} respectively)"
             .format(mean=mean, stdev=stdev, thresh_mean=mean_threshold,
                     thresh_stdev=stdev_threshold, a=out_path, b=ref_path)))

    def get_session_dir(self, subject=None, visit=None,
                        frequency='per_session'):
        if subject is None and frequency in ('per_session', 'per_subject'):
            subject = self.SUBJECT
        if visit is None and frequency in ('per_session', 'per_visit'):
            visit = self.VISIT
        if frequency == 'per_session':
            assert subject is not None
            assert visit is not None
            path = op.join(self.project_dir, subject, visit)
        elif frequency == 'per_subject':
            assert subject is not None
            assert visit is None
            path = op.join(
                self.project_dir, subject, SUMMARY_NAME)
        elif frequency == 'per_visit':
            assert visit is not None
            assert subject is None
            path = op.join(self.project_dir, SUMMARY_NAME, visit)
        elif frequency == 'per_project':
            assert subject is None
            assert visit is None
            path = op.join(self.project_dir, SUMMARY_NAME, SUMMARY_NAME)
        else:
            assert False
        return op.abspath(path)

    @classmethod
    def remove_generated_files(cls, study=None):
        # Remove derived datasets
        for fname in os.listdir(cls.get_session_dir()):
            if study is None or fname.startswith(study + '_'):
                os.remove(op.join(cls.get_session_dir(), fname))

    def output_file_path(self, fname, study_name, subject=None, visit=None,
                         frequency='per_session', **kwargs):
        return op.join(
            self.get_session_dir(subject=subject, visit=visit,
                                 frequency=frequency, **kwargs),
            '{}_{}'.format(study_name, fname))

    def ref_file_path(self, fname, subject=None, session=None):
        return op.join(self.session_dir, fname,
                            subject=subject, session=session)


class BaseMultiSubjectTestCase(BaseTestCase):

    SUMMARY_NAME = LOCAL_SUMMARY_NAME

    def setUp(self):
        self.reset_dirs()
        self.add_sessions(self.project_dir)

    def add_sessions(self, project_dir):
        self.local_tree = deepcopy(self.input_tree)
        if project_dir is not None:  # For local archive
            proj_summ_path = op.join(project_dir, SUMMARY_NAME,
                                     SUMMARY_NAME)
            for dataset in self.local_tree.datasets:
                self.init_dataset(dataset, op.join(proj_summ_path,
                                                   dataset.fname()))
            self.init_fields(proj_summ_path, self.local_tree.fields)
            for visit in self.local_tree.visits:
                visit_summ_path = op.join(project_dir, SUMMARY_NAME,
                                          visit.id)
                for dataset in visit.datasets:
                    self.init_dataset(dataset, op.join(visit_summ_path,
                                                       dataset.fname()))
                self.init_fields(visit_summ_path, visit.fields)
            for subject in self.local_tree.subjects:
                subj_summ_path = op.join(project_dir, subject.id,
                                         SUMMARY_NAME)
                for dataset in subject.datasets:
                    self.init_dataset(dataset, op.join(subj_summ_path,
                                                       dataset.fname()))
                self.init_fields(subj_summ_path, subject.fields)
                for session in subject.sessions:
                    sess_path = op.join(project_dir, session.subject_id,
                                        session.visit_id)
                    for dataset in session.datasets:
                        self.init_dataset(
                            dataset, op.join(sess_path,
                                             dataset.fname()))
                    self.init_fields(sess_path, session.fields)

    @property
    def subject_ids(self):
        return (d for d in os.listdir(self.project_dir)
                if d != self.SUMMARY_NAME)

    def visit_ids(self, subject_id):
        subject_dir = op.join(self.project_dir, subject_id)
        return (d for d in os.listdir(subject_dir)
                if d != self.SUMMARY_NAME)

    def session_dir(self, subject, visit):
        return self.get_session_dir(subject, visit)

    def get_session_dir(self, subject, visit, frequency='per_session'):
        return super(BaseMultiSubjectTestCase, self).get_session_dir(
            subject=subject, visit=visit, frequency=frequency)

    def init_dataset(self, dataset, path):
        dataset._path = path
        dataset._archive = self.archive
        self._make_dir(op.dirname(dataset.path))
        with open(dataset.path, 'w') as f:
            f.write(str(self.DATASET_CONTENTS[dataset.name]))

    def init_fields(self, dpath, fields):
        self._make_dir(dpath)
        dct = {f.name: f.value for f in fields}
        with open(op.join(dpath, FIELDS_FNAME), 'w') as f:
            json.dump(dct, f)

    def _make_dir(self, path):
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return path


class DummyTestCase(BaseTestCase):

    def __init__(self):
        self.setUp()

    def __del__(self):
        self.tearDown()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def assert_(self, statement, message=None):
        if not statement:
            message = "'{}' is not true".format(statement)
            print message
        else:
            print "Test successful"

    def assertEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print message
        else:
            print "Test successful"

    def assertAlmostEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print message
        else:
            print "Test successful"

    def assertLess(self, first, second, message=None):
        if first >= second:
            if message is None:
                message = '{} is not less than {}'.format(repr(first),
                                                          repr(second))
            print message
        else:
            print "Test successful"


class TestTestCase(BaseTestCase):

    def test_test(self):
        pass


class TestMathInputSpec(TraitedSpec):

    x = traits.Either(traits.Float(), traits.File(exists=True),
                      traits.List(traits.Float),
                      traits.List(traits.File(exists=True)),
                      mandatory=True, desc='first arg')
    y = traits.Either(traits.Float(), traits.File(exists=True),
                      mandatory=False, desc='second arg')
    op = traits.Str(mandatory=True, desc='operation')

    z = traits.File(genfile=True, mandatory=False,
                    desc="Name for output file")

    as_file = traits.Bool(False, desc="Whether to write as a file",
                          usedefault=True)


class TestMathOutputSpec(TraitedSpec):

    z = traits.Either(traits.Float(), traits.File(exists=True),
                      'output')


class TestMath(BaseInterface):
    """
    A basic interface to test out the pipeline infrastructure
    """

    input_spec = TestMathInputSpec
    output_spec = TestMathOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        x = self.inputs.x
        y = self.inputs.y
        if isinstance(x, basestring):
            x = self._load_file(x)
        if isinstance(y, basestring):
            y = self._load_file(y)
        oper = getattr(operator, self.inputs.op)
        if isdefined(y):
            z = oper(x, y)
        elif isinstance(x, list):
            if isinstance(x[0], basestring):
                x = [self._load_file(u) for u in x]
            z = reduce(oper, x)
        else:
            raise Exception(
                "If 'y' is not provided then x needs to be list")
        outputs = self.output_spec().get()
        if self.inputs.as_file:
            z_path = op.abspath(self._gen_z_fname())
            with open(z_path, 'w') as f:
                f.write(str(z))
            outputs['z'] = z_path
        else:
            outputs['z'] = z
        return outputs

    def _gen_filename(self, name):
        if name == 'z':
            fname = self._gen_z_fname()
        else:
            assert False
        return fname

    def _gen_z_fname(self):
        if isdefined(self.inputs.z):
            fname = self.inputs.z
        else:
            fname = 'z.txt'
        return fname

    @classmethod
    def _load_file(self, path):
        with open(path) as f:
            return float(f.read())



from __future__ import absolute_import
import os.path
import re
import shutil
from nipype.interfaces.base import (
    TraitedSpec, traits, BaseInterface, File, isdefined,
    Directory, CommandLineInputSpec, CommandLine, DynamicTraitedSpec,
    BaseInterfaceInputSpec)
from arcana.exception import ArcanaUsageError
from itertools import chain
from nipype.interfaces.utility.base import Merge, MergeInputSpec
from nipype.interfaces.utility import IdentityInterface
from nipype.interfaces.io import IOBase, add_traits
from nipype.utils.filemanip import filename_to_list
from nipype.interfaces.base import OutputMultiPath, InputMultiPath
import numpy as np


zip_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'zip.sh'))
targz_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'targz.sh'))
cp_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'copy_file.sh'))
cp_dir_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'copy_dir.sh'))
mkdir_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'resources', 'bash',
                 'make_dir.sh'))


special_char_re = re.compile(r'[^\w]')


class MergeInputSpec(DynamicTraitedSpec, BaseInterfaceInputSpec):
    axis = traits.Enum(
        'vstack', 'hstack', usedefault=True,
        desc=('direction in which to merge, hstack requires same number '
              'of elements in each input'))
    no_flatten = traits.Bool(
        False, usedefault=True,
        desc='append to outlist instead of extending in vstack mode')


class MergeOutputSpec(TraitedSpec):
    out = traits.List(desc='Merged output')


class Merge(IOBase):
    """Basic interface class to merge inputs into a single list

    Examples
    --------

    >>> from nipype.interfaces.utility import Merge
    >>> mi = Merge(3)
    >>> mi.inputs.in1 = 1
    >>> mi.inputs.in2 = [2, 5]
    >>> mi.inputs.in3 = 3
    >>> out = mi.run()
    >>> out.outputs.out
    [1, 2, 5, 3]

    """
    input_spec = MergeInputSpec
    output_spec = MergeOutputSpec

    def __init__(self, numinputs=0, **inputs):
        super(Merge, self).__init__(**inputs)
        self._numinputs = numinputs
        if numinputs > 0:
            input_names = ['in%d' % (i + 1) for i in range(numinputs)]
        elif numinputs == 0:
            input_names = ['in_lists']
        else:
            input_names = []
        add_traits(self.inputs, input_names)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out = []

        if self._numinputs == 0:
            values = getattr(self.inputs, 'in_lists')
            if not isdefined(values):
                return outputs
        else:
            getval = lambda idx: getattr(self.inputs, 'in%d' % (idx + 1))  # @IgnorePep8
            values = [getval(idx) for idx in range(self._numinputs)
                      if isdefined(getval(idx))]

        if self.inputs.axis == 'vstack':
            for value in values:
                if isinstance(value, list) and not self.inputs.no_flatten:
                    out.extend(value)
                else:
                    out.append(value)
        else:
            lists = [filename_to_list(val) for val in values]
            out = [[val[i] for val in lists] for i in range(len(lists[0]))]
        if out:
            outputs['out'] = out
        return outputs


class MergeTupleOutputSpec(TraitedSpec):
    out = traits.Tuple(desc='Merged output')  # @UndefinedVariable


class MergeTuple(Merge):
    """Basic interface class to merge inputs into a single tuple

    Examples
    --------

    >>> from nipype.interfaces.utility import Merge
    >>> mi = MergeTuple(3)
    >>> mi.inputs.in1 = 1
    >>> mi.inputs.in2 = [2, 5]
    >>> mi.inputs.in3 = 3
    >>> out = mi.run()
    >>> out.outputs.out
    (1, 2, 5, 3)

    """
    input_spec = MergeInputSpec
    output_spec = MergeTupleOutputSpec

    def _list_outputs(self):
        outputs = super(MergeTuple, self)._list_outputs()
        outputs['out'] = tuple(outputs['out'])
        return outputs


class Chain(IdentityInterface):

    def _list_outputs(self):
        outputs = super(Chain, self)._list_outputs()
        chained_outputs = {}
        for k, v in outputs.iteritems():
            chained_outputs[k] = list(chain(*v))
        return chained_outputs


def dicom_fname_sort_key(fname):
    in_parts = special_char_re.split(os.path.basename(fname))
    out_parts = []
    for part in in_parts:
        try:
            part = int(part)
        except ValueError:
            pass
        out_parts.append(part)
    return tuple(out_parts)


class JoinPathInputSpec(TraitedSpec):
    dirname = Directory(mandatory=True, desc='directory name')
    filename = traits.Str(mandatory=True, desc='file name')


class JoinPathOutputSpec(TraitedSpec):
    path = traits.Str(mandatory=True, desc="The joined path")


class JoinPath(BaseInterface):
    """Joins a filename to a directory name"""

    input_spec = JoinPathInputSpec
    output_spec = JoinPathOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['path'] = os.path.join(self.inputs.dirname,
                                       self.inputs.filename)
        return outputs

    def _run_interface(self, runtime):
        return runtime


class CopyFileInputSpec(CommandLineInputSpec):
    src = File(mandatory=True, desc='source file', argstr='%s',
               position=0)
    base_dir = Directory(mandatory=True, desc='root directory', argstr='%s',
                         position=1)
    dst = File(genfile=True, argstr='%s', position=2,
               desc=("The destination file"))


class CopyFileOutputSpec(TraitedSpec):
    copied = File(exists=True, desc="The copied file")
    basedir = Directory(exists=True, desc='base directory')


class CopyFile(CommandLine):
    """Creates a copy of a given file"""

    _cmd = cp_path
    input_spec = CopyFileInputSpec
    output_spec = CopyFileOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs['copied'] = os.path.join(self.inputs.base_dir, self.inputs.dst)
        outputs['basedir'] = os.path.join(self.inputs.base_dir)
        return outputs

    def _gen_filename(self, name):
        if name == 'copied':
            fname = os.path.basename(self.inputs.dst)
        else:
            assert False
        return fname


class CopyDirInputSpec(CommandLineInputSpec):
    src = File(mandatory=True, desc='source file', argstr='%s',
               position=0)
    base_dir = Directory(mandatory=True, desc='root directory', argstr='%s',
                         position=1)
    dst = File(genfile=True, argstr='%s', position=2,
               desc=("The destination file"))
    method = traits.Int(mandatory=True, desc='method', argstr='%s', position=3)


class CopyDirOutputSpec(TraitedSpec):
    copied = Directory(exists=True, desc="The copied file")
    basedir = Directory(exists=True, desc='base directory')


class CopyDir(CommandLine):
    """Creates a copy of a given file"""

    _cmd = cp_dir_path
    input_spec = CopyDirInputSpec
    output_spec = CopyDirOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        if self.inputs.method == 1:
            outputs['copied'] = os.path.join(self.inputs.base_dir)
            outputs['basedir'] = os.path.join(self.inputs.base_dir)
        elif self.inputs.method == 2:
            outputs['copied'] = os.path.join(self.inputs.base_dir,
                                             self._gen_filename('copied'))
            outputs['basedir'] = os.path.join(self.inputs.base_dir)
        return outputs

    def _gen_filename(self, name):
        if name == 'copied':
            fname = os.path.basename(self.inputs.dst)
        else:
            assert False
        return fname


class MakeDirInputSpec(CommandLineInputSpec):
    base_dir = Directory(mandatory=True, desc='root directory', argstr='%s',
                         position=0)
    name_dir = Directory(genfile=True, argstr='%s', position=1,
                         desc=("name of the new directory"))


class MakeDirOutputSpec(TraitedSpec):
    new_dir = Directory(exists=True, desc="The created directory")


class MakeDir(CommandLine):
    """Creates a new directory"""

    _cmd = mkdir_path
    input_spec = MakeDirInputSpec
    output_spec = MakeDirOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['new_dir'] = os.path.join(self.inputs.base_dir)
#                                           self._gen_filename('new_dir'))
        return outputs

    def _gen_filename(self, name):
        if name == 'new_dir':
            fname = os.path.basename(self.inputs.name_dir)
        else:
            assert False
        return fname


class ZipDirInputSpec(CommandLineInputSpec):
    dirname = Directory(mandatory=True, desc='directory name', argstr='%s',
                        position=1)
    zipped = File(genfile=True, argstr='%s', position=0,
                  desc=("The zipped zip file"))
    ext_prefix = traits.Str(
        mandatory=False, default='', usedefault=True,
        desc=("Extra extension to prepend before .zip is appended to "
              "file name"))


class ZipDirOutputSpec(TraitedSpec):
    zipped = File(exists=True, desc="The zipped directory")


class ZipDir(CommandLine):
    """Creates a zip archive from a given folder"""

    _cmd = zip_path
    input_spec = ZipDirInputSpec
    output_spec = ZipDirOutputSpec
    zip_ext = '.zip'

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['zipped'] = os.path.abspath(
            self._gen_filename('zipped'))
        return outputs

    def _gen_filename(self, name):
        if name == 'zipped':
            if isdefined(self.inputs.zipped):
                fname = self.inputs.zipped
            else:
                fname = (os.path.basename(self.inputs.dirname) +
                         self.inputs.ext_prefix + self.zip_ext)
        else:
            assert False
        return fname


class UnzipDirInputSpec(CommandLineInputSpec):
    zipped = Directory(mandatory=True, desc='zipped file name', argstr='%s',
                       position=0)


class UnzipDirOutputSpec(TraitedSpec):
    unzipped = Directory(exists=True, desc="The unzipped directory")


class UnzipDir(CommandLine):
    """Unzips a folder that was zipped by ZipDir"""

    _cmd = 'unzip -qo'
    input_spec = UnzipDirInputSpec
    output_spec = UnzipDirOutputSpec

    def _run_interface(self, *args, **kwargs):
        self.listdir_before = set(os.listdir(os.getcwd()))
        return super(UnzipDir, self)._run_interface(*args, **kwargs)

    def _list_outputs(self):
        outputs = self._outputs().get()
        new_files = set(os.listdir(os.getcwd())) - self.listdir_before
        if len(new_files) > 1:
            raise ArcanaUsageError(
                "Zip archives can only contain a single directory, found '{}'"
                .format("', '".join(new_files)))
        try:
            unzipped = next(iter(new_files))
        except StopIteration:
            raise ArcanaUsageError(
                "No files or directories found in unzipped directory")
        outputs['unzipped'] = os.path.join(os.getcwd(), unzipped)
        return outputs


class CopyToDirInputSpec(TraitedSpec):
    in_files = traits.List(traits.Either(File(exists=True),
                                         Directory(exists=True)), mandatory=True,
                           desc='input dicom files')
    out_dir = File(genfile=True, desc='the output dicom file')
    extension = traits.Str(desc='specify the extention for the copied file.',
                           default='', usedefault=True)


class CopyToDirOutputSpec(TraitedSpec):
    out_dir = Directory(exists=True, desc='the output dicom directory')


class CopyToDir(BaseInterface):
    """
    Copies a list of files of directories into a directory
    """

    input_spec = CopyToDirInputSpec
    output_spec = CopyToDirOutputSpec

    def _run_interface(self, runtime):
        dirname = self._gen_outdirname()
        os.makedirs(dirname)
        ext = self.inputs.extension
        for i, f in enumerate(self.inputs.in_files):
            if os.path.isdir(f):
                out_name = f.split('/')[-1]
                if ext:
                    out_name = '{0}_{1}'.format(
                        out_name, ext+str(i).zfill(3))
                shutil.copytree(f, dirname+'/{}'.format(out_name))
            elif os.path.isfile(f):
                if ext == '.dcm':
                    fname = os.path.join(dirname, str(i).zfill(4)) + ext
                else:
                    fname = dirname
                shutil.copy(f, fname)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_dir'] = self._gen_outdirname()
        return outputs

    def _gen_filename(self, name):
        if name == 'out_dir':
            fname = self._gen_outdirname()
        else:
            assert False
        return fname

    def _gen_outdirname(self):
        if isdefined(self.inputs.out_dir):
            dpath = self.inputs.out_dir
        else:
            dpath = os.path.join(os.getcwd(), 'store_dir')
        return dpath


class ListDirInputSpec(TraitedSpec):
    directory = File(mandatory=True, desc='directory to read')
    sort_key = traits.Function(
        desc=("A function that generates a key from the listed filenames with "
              "which to sort them with"))


class ListDirOutputSpec(TraitedSpec):
    files = traits.List(File(exists=True),
                        desc='The files present in the directory')


class ListDir(BaseInterface):
    """
    Lists all files (not sub-directories) in a directory
    """

    input_spec = ListDirInputSpec
    output_spec = ListDirOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        dname = self.inputs.directory
        outputs = self._outputs().get()
        key = self.inputs.sort_key if isdefined(self.inputs.sort_key) else None
        outputs['files'] = sorted((os.path.join(dname, f)
                                   for f in os.listdir(dname)
                                   if os.path.isfile(os.path.join(dname, f))),
                                  key=key)
        return outputs

class TarGzDirInputSpec(CommandLineInputSpec):
    dirname = Directory(mandatory=True, desc='directory name', argstr='%s',
                        position=1)
    zipped = File(genfile=True, argstr='%s', position=0,
                  desc=("The tar_gz file"))


class TarGzDirOutputSpec(TraitedSpec):
    zipped = File(exists=True, desc="The tar_gz directory")


class TarGzDir(CommandLine):
    """Creates a tar_gzip archive from a given folder"""

    _cmd = targz_path
    input_spec = TarGzDirInputSpec
    output_spec = TarGzDirOutputSpec
    targz_ext = '.tar.gz'

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['zipped'] = os.path.join(os.getcwd(),
                                         self._gen_filename('zipped'))
        return outputs

    def _gen_filename(self, name):
        if name == 'zipped':
            fname = os.path.basename(self.inputs.dirname) + self.targz_ext
        else:
            assert False
        return fname


class UnTarGzDirInputSpec(CommandLineInputSpec):

    gzipped = Directory(mandatory=True, argstr='%s', position=0,
                        desc=("The tar_gz file"))


class UnTarGzDirOutputSpec(TraitedSpec):
    gunzipped = Directory(exists=True, desc="The gunzipped directory")


class UnTarGzDir(CommandLine):
    """Unzip a folder created using TarGz"""

    _cmd = 'tar -zxvf '
    input_spec = UnTarGzDirInputSpec
    output_spec = UnTarGzDirOutputSpec

    def _run_interface(self, *args, **kwargs):
        self.listdir_before = set(os.listdir(os.getcwd()))
        return super(UnTarGzDir, self)._run_interface(*args, **kwargs)

    def _list_outputs(self):
        outputs = self._outputs().get()
        new_files = set(os.listdir(os.getcwd())) - self.listdir_before
        if len(new_files) > 1:
            raise ArcanaUsageError(
                "Zip archives can only contain a single directory, found '{}'"
                .format("', '".join(new_files)))
        try:
            unzipped = next(iter(new_files))
        except StopIteration:
            raise ArcanaUsageError(
                "No files or directories found in unzipped directory")
        outputs['gunzipped'] = os.path.join(os.getcwd(), unzipped)
        return outputs


class SelectOneInputSpec(BaseInterfaceInputSpec):
    inlist = InputMultiPath(
        traits.Any, mandatory=True, desc='list of values to choose from')
    index = traits.Int(mandatory=True, desc='0-based indice of element to extract')


class SelectOneOutputSpec(TraitedSpec):
    out = OutputMultiPath(traits.Any, desc='selected value')


class SelectOne(IOBase):
    """Basic interface class to select specific elements from a list

    Examples
    --------

    >>> from nipype.interfaces.utility import Select
    >>> sl = Select()
    >>> _ = sl.inputs.trait_set(inlist=[1, 2, 3, 4, 5], index=[3])
    >>> out = sl.run()
    >>> out.outputs.out
    4

    >>> _ = sl.inputs.trait_set(inlist=[1, 2, 3, 4, 5], index=[3, 4])
    >>> out = sl.run()
    >>> out.outputs.out
    [4, 5]

    """

    input_spec = SelectOneInputSpec
    output_spec = SelectOneOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        out = np.array(self.inputs.inlist)[self.inputs.index]
        outputs['out'] = out
        return outputs



import re
from itertools import chain
from copy import copy
from collections import defaultdict
from arcana.exception import (
    ArcanaError, ArcanaUsageError,
    ArcanaDatasetMatchError)
from .base import BaseDataset, BaseField


class BaseMatch(object):
    """
    Base class for Dataset and Field Match classes
    """

    def __init__(self, pattern, derived, order, study, is_regex):
        self._derived = derived
        self._order = order
        self._pattern = pattern
        self._study = study
        self._is_regex = is_regex
        self._matches = None

    def __eq__(self, other):
        return (self.derived == other.derived and
                self.pattern == other.pattern and
                self.order == other.order and
                self._study == other._study)

    @property
    def pattern(self):
        return self._pattern

    @property
    def derived(self):
        return self._derived

    @property
    def study(self):
        return self._study

    @property
    def matches(self):
        if self.frequency == 'per_session':
            matches = chain(*(d.itervalues()
                              for d in self._matches.itervalues()))
        elif self.frequency == 'per_subject':
            matches = self._matches.itervalues()
        elif self.frequency == 'per_visit':
            matches = self._matches.itervalues()
        elif self.frequency == 'per_project':
            self._matches = iter([self._matches])
        else:
            assert False
        return matches

    @property
    def is_regex(self):
        return self._is_regex

    @property
    def order(self):
        return self._order

    def bind(self, study):
        if self._study is not None:
            bound = self
        else:
            bound = copy(self)
            bound._study = study
            bound._match_tree(study.tree)
        return bound

    @property
    def prefixed_name(self):
        return self.name

    def basename(self, **kwargs):
        if not self.is_regex:
            basename = self.pattern
        else:
            basename = self.match(**kwargs).name
        return basename

    def _match_tree(self, tree, **kwargs):
        # Run the match against the tree
        if self.frequency == 'per_session':
            self._matches = defaultdict(dict)
            for subject in tree.subjects:
                for sess in subject.sessions:
                    if self.derived and sess.derived is not None:
                        node = sess.derived
                    else:
                        node = sess
                    self._matches[sess.subject_id][
                        sess.visit_id] = self._match_node(node,
                                                          **kwargs)
        elif self.frequency == 'per_subject':
            self._matches = {}
            for subject in tree.subjects:
                self._matches[subject.id] = self._match_node(subject,
                                                             **kwargs)
        elif self.frequency == 'per_visit':
            self._matches = {}
            for visit in tree.visits:
                self._matches[visit.id] = self._match_node(visit,
                                                           **kwargs)
        elif self.frequency == 'per_project':
            self._matches = self._match_node(tree, **kwargs)
        else:
            assert False, "Unrecognised frequency '{}'".format(
                self.frequency)

    def _match_node(self, node, **kwargs):
        # Get names matching pattern
        matches = self._filtered_matches(node, **kwargs)
        # Select the dataset from the matches
        if self.order is not None:
            try:
                match = matches[self.order]
            except IndexError:
                raise ArcanaDatasetMatchError(
                    "Did not find {} datasets names matching pattern {}"
                    " (found {}) in {}".format(self.order, self.pattern,
                                               len(matches), node))
        elif len(matches) == 1:
            match = matches[0]
        elif matches:
            raise ArcanaDatasetMatchError(
                "Found multiple matches for {} pattern in {} ({})"
                .format(self.pattern, node,
                        ', '.join(str(m) for m in matches)))
        else:
            raise ArcanaDatasetMatchError(
                "Did not find any matches for {} pattern in {} "
                "(found {})"
                .format(self.pattern, node,
                        ', '.join(str(d) for d in node.datasets)))
        return match

    def match(self, subject_id=None, visit_id=None):
        if self._matches is None:
            raise ArcanaError(
                "{} has not been bound to study".format(self))
        if self.frequency == 'per_session':
            if subject_id is None or visit_id is None:
                raise ArcanaError(
                    "The 'subject_id' and 'visit_id' must be provided "
                    "to get the match from {}".format(self))
            dataset = self._matches[subject_id][visit_id]
        elif self.frequency == 'per_subject':
            if subject_id is None:
                raise ArcanaError(
                    "The 'subject_id' arg must be provided to get "
                    "the match from {}"
                    .format(self))
            dataset = self._matches[subject_id]
        elif self.frequency == 'per_visit':
            if visit_id is None:
                raise ArcanaError(
                    "The 'visit_id' arg must be provided to get "
                    "the match from {}"
                    .format(self))
            dataset = self._matches[visit_id]
        elif self.frequency == 'per_project':
            dataset = self._matches
        return dataset

    def initkwargs(self):
        dct = {}
        dct['derived'] = self.derived
        dct['study'] = self._study
        dct['pattern'] = self.pattern
        dct['order'] = self.order
        return dct


class DatasetMatch(BaseDataset, BaseMatch):
    """
    A pattern that describes a single dataset (typically acquired
    rather than generated but not necessarily) within each session.

    Parameters
    ----------
    name : str
        The name of the dataset, typically left None and set in Study
    pattern : str
        A regex pattern to match the dataset names with. Must match
        one and only one dataset per <frequency>. If None, the name
        is used instead.
    format : FileFormat
        The file format used to store the dataset. Can be one of the
        recognised formats
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_project',
        specifying whether the dataset is present for each session, subject,
        visit or project.
    derived : bool
        Whether the scan was generated or acquired. Depending on the archive
        used to store the dataset this is used to determine the location of the
        dataset.
    id : int | None
        To be used to distinguish multiple datasets that match the
        pattern in the same session. The ID of the dataset within the
        session.
    order : int | None
        To be used to distinguish multiple datasets that match the
        pattern in the same session. The order of the dataset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    dicom_tags : dct(str | str)
        To be used to distinguish multiple datasets that match the
        pattern in the same session. The provided DICOM values dicom
        header values must match exactly.
    """

    is_spec = False

    def __init__(self, name, format, pattern=None, # @ReservedAssignment @IgnorePep8
                 frequency='per_session', derived=False, id=None,  # @ReservedAssignment @IgnorePep8
                 order=None, dicom_tags=None, is_regex=False,
                 study=None):
        if pattern is None and id is None:
            raise ArcanaUsageError(
                "Either 'pattern' or 'id' need to be provided to "
                "DatasetMatch constructor")
        BaseDataset.__init__(self, name, format, frequency)
        BaseMatch.__init__(self, pattern, derived, order, study,
                           is_regex)
        if dicom_tags is not None and format.name != 'dicom':
            raise ArcanaUsageError(
                "Cannot use 'dicom_tags' kwarg with non-DICOM "
                "format ({})".format(format))
        self._dicom_tags = dicom_tags
        if order is not None and id is not None:
            raise ArcanaUsageError(
                "Cannot provide both 'order' and 'id' to a dataset"
                "match")
        self._id = str(id) if id is not None else id

    def __eq__(self, other):
        return (BaseDataset.__eq__(self, other) and
                BaseMatch.__eq__(self, other) and
                self.dicom_tags == other.dicom_tags and
                self.id == other.id)

    def initkwargs(self):
        dct = BaseDataset.initkwargs(self)
        dct.update(BaseMatch.initkwargs(self))
        dct['dicom_tags'] = self.dicom_tags
        dct['id'] = self.id
        return dct

    def __repr__(self):
        return ("{}(name='{}', format={}, frequency={}, derived={},"
                " pattern={}, order={}, id={}, dicom_tags={}, "
                "is_regex={}, study={})"
                .format(self.__class__.__name__, self.name, self.format,
                        self.frequency, self.derived, self._pattern,
                        self.order, self.id, self.dicom_tags,
                        self.is_regex, self._study))

    @property
    def id(self):
        return self._id

    @property
    def dicom_tags(self):
        return self._dicom_tags

    def _match_tree(self, tree, **kwargs):
        with self.study.archive.login() as archive_login:
            super(DatasetMatch, self)._match_tree(
                tree, archive_login=archive_login, **kwargs)

    def _filtered_matches(self, node, archive_login=None):
        if self.pattern is not None:
            if self.is_regex:
                pattern_re = re.compile(self.pattern)
                matches = [d for d in node.datasets
                           if pattern_re.match(d.name)]
            else:
                matches = [d for d in node.datasets
                           if d.name == self.pattern]
        else:
            matches = list(node.datasets)
        if not matches:
            raise ArcanaDatasetMatchError(
                "No dataset names in {} match '{}' pattern:{}"
                .format(node, self.pattern,
                        '\n'.join(d.name for d in node.datasets)))
        if self.id is not None:
            filtered = [d for d in matches if d.id == self.id]
            if not filtered:
                raise ArcanaDatasetMatchError(
                    "Did not find datasets names matching pattern {}"
                    "with an id of {} (found {}) in {}".format(
                        self.pattern, self.id,
                        ', '.join(str(m) for m in matches), node))
            matches = filtered
        # Filter matches by dicom tags
        if self.dicom_tags is not None:
            filtered = []
            for dataset in matches:
                values = dataset.dicom_values(
                    self.dicom_tags.keys(), archive_login=archive_login)
                if self.dicom_tags == values:
                    filtered.append(dataset)
            if not filtered:
                raise ArcanaDatasetMatchError(
                    "Did not find datasets names matching pattern {}"
                    "that matched DICOM tags {} (found {}) in {}"
                    .format(self.pattern, self.dicom_tags,
                            ', '.join(str(m) for m in matches), node))
            matches = filtered
        return matches


class FieldMatch(BaseField, BaseMatch):
    """
    A pattern that matches a single field (typically acquired rather than
    generated but not necessarily) in each session.

    Parameters
    ----------
    name : str
        The name of the dataset
    pattern : str
        A regex pattern to match the field names with. Must match
        one and only one dataset per <frequency>. If None, the name
        is used instead.
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_project',
        specifying whether the dataset is present for each session, subject,
        visit or project.
    derived : bool
        Whether or not the value belongs in the derived session or not
    order : int | None
        To be used to distinguish multiple datasets that match the
        pattern in the same session. The order of the dataset within the
        session. Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    """

    is_spec = False

    def __init__(self, name, dtype, pattern, frequency='per_session',
                 derived=False, order=None, is_regex=False, study=None):
        BaseField.__init__(self, name, dtype, frequency)
        BaseMatch.__init__(self, pattern, derived, order, study,
                           is_regex)
        super(FieldMatch, self).__init__(name, dtype, frequency)

    def __eq__(self, other):
        return (BaseField.__eq__(self, other) and
                BaseMatch.__eq__(self, other))

    def initkwargs(self):
        dct = BaseField.initkwargs(self)
        dct.update(BaseMatch.initkwargs(self))
        return dct

    def _filtered_matches(self, node):
        if self.is_regex:
            pattern_re = re.compile(self.pattern)
            matches = [f for f in node.fields
                       if pattern_re.match(f.name)]
        else:
            matches = [f for f in node.fields
                       if f.name == self.pattern]
        if not matches:
            raise ArcanaDatasetMatchError(
                "No field names in {} match '{}' pattern"
                .format(node, self.pattern))
        return matches

    def __repr__(self):
        return ("{}(name='{}', dtype={}, frequency={}, derived={},"
                " pattern={}, order={}, is_regex={}, study={})"
                .format(self.__class__.__name__, self.name, self.dtype,
                        self.frequency, self.derived,
                        self._pattern, self.order, self.is_regex,
                        self._study))



from itertools import chain
from nipype.interfaces.base import TraitedSpec, traits, BaseInterface
from arcana.exception import ArcanaError


class InputSubjectsInputSpec(TraitedSpec):

    subject_id = traits.Str(mandatory=True, desc=("The subject ID"))
    prereq_reports = traits.List(
        traits.Tuple(traits.Str, traits.Str), value=[],
        desc=("A list of session reports from all prerequisite pipelines"))


class InputSubjectsOutputSpec(TraitedSpec):

    subject_id = traits.Str(mandatory=True, desc=("The subject ID"))


class InputSubjects(BaseInterface):
    """
    Basically an IndentityInterface for iterating over all subjects with an
    extra input to feed the output summaries of prerequisite pipelines into to
    make sure they are run before the current pipeline.
    """

    input_spec = InputSubjectsInputSpec
    output_spec = InputSubjectsOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['subject_id'] = self.inputs.subject_id
        return outputs


class InputSessionsSpec(TraitedSpec):

    visit_id = traits.Str(mandatory=True, desc=("The visit ID"))
    subject_id = traits.Str(mandatory=True, desc=("The subject ID"))


class InputSessions(BaseInterface):
    """
    Basically an IndentityInterface for iterating over all session.
    """

    input_spec = InputSessionsSpec
    output_spec = InputSessionsSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['visit_id'] = self.inputs.visit_id
        outputs['subject_id'] = self.inputs.subject_id
        return outputs


class SessionReportInputSpec(TraitedSpec):

    sessions = traits.List(traits.Str)
    subjects = traits.List(traits.Str)


class SessionReportOutputSpec(TraitedSpec):

    subject_session_pairs = traits.List(traits.Tuple(traits.Str, traits.Str))


class SessionReport(BaseInterface):
    """
    Basically an IndentityInterface for joining over sessions
    """

    input_spec = SessionReportInputSpec
    output_spec = SessionReportOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['subject_session_pairs'] = list(zip(self.inputs.subjects,
                                                    self.inputs.sessions))
        return outputs


class SubjectReportSpec(TraitedSpec):

    subjects = traits.List(traits.Str)


class SubjectReport(BaseInterface):
    """
    Basically an IndentityInterface for joining over subjects
    """

    input_spec = SubjectReportSpec
    output_spec = SubjectReportSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['subjects'] = self.inputs.subjects
        return outputs


class VisitReportSpec(TraitedSpec):

    sessions = traits.List(traits.Str)


class VisitReport(BaseInterface):
    """
    Basically an IndentityInterface for joining over sessions
    """

    input_spec = VisitReportSpec
    output_spec = VisitReportSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['sessions'] = self.inputs.sessions
        return outputs


class SubjectSessionReportInputSpec(TraitedSpec):

    subject_session_pairs = traits.List(
        traits.List(traits.Tuple(traits.Str, traits.Str)))


class SubjectSessionReportOutputSpec(TraitedSpec):

    subject_session_pairs = traits.List(traits.Tuple(traits.Str, traits.Str))


class SubjectSessionReport(BaseInterface):
    """
    Basically an IndentityInterface for joining over subject-session pairs
    """

    input_spec = SubjectSessionReportInputSpec
    output_spec = SubjectSessionReportOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['subject_session_pairs'] = list(
            chain(*self.inputs.subject_session_pairs))
        return outputs


class PipelineReportInputSpec(TraitedSpec):
    subject_session_pairs = traits.List(traits.Tuple(
        traits.Str, traits.Str),
        desc="Subject & session pairs from per-session sink")
    subjects = traits.List(traits.Str,
                           desc="Subjects from per-subject sink")
    visits = traits.List(traits.Str,
                             desc="Visits from per_visit sink")
    project = traits.Str(desc="Project ID from per-project sink")


class PipelineReportOutputSpec(TraitedSpec):
    subject_session_pairs = traits.List(traits.Tuple(
        traits.Str, traits.Str),
        desc="Session & subject pairs from per-session sink")
    subjects = traits.List(traits.Str, desc="Subjects from per-subject sink")
    project = traits.Str(desc="Project ID from per-project sink")


class PipelineReport(BaseInterface):

    input_spec = PipelineReportInputSpec
    output_spec = PipelineReportOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        outputs['subject_session_pairs'] = self.inputs.subject_session_pairs
        outputs['subjects'] = self.inputs.subjects
        outputs['project'] = self.inputs.project
        return outputs


class SelectSessionInputSpec(TraitedSpec):
    subject_id = traits.Str(mandatory=True, desc="The subject ID to select")
    visit_id = traits.Str(mandatory=True, desc="The visit ID to select")
    subject_ids = traits.List(
        traits.Str(), mandatory=True,
        desc="The subject IDs to select the subject_id from")
    visit_ids = traits.List(
        traits.Str(), mandatory=True,
        desc="The visit IDs to select the visit_id from")
    items = traits.List(
        traits.Any(),
        mandatory=True,
        desc=("The items from which to select the one corresponding to the "
              "session from"))


class SelectSessionOutputSpec(TraitedSpec):
    item = traits.Any(desc="The selected item")


class SelectSession(BaseInterface):

    input_spec = SelectSessionInputSpec
    output_spec = SelectSessionOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = {}
        session_id = (self.inputs.subject_id, self.inputs.visit_id)
        session_ids = zip(self.inputs.subject_ids, self.inputs.visit_ids)
        if session_ids.count(session_id) != 1:
            raise ArcanaError(
                "More than one indices matched {} in subjects and visits list "
                "({})".format(session_id, session_ids))
        index = session_ids.index(session_id)
        outputs['item'] = self.inputs.items[index]
        return outputs



from .base import BaseRunner
from nipype.pipeline.plugins import LinearPlugin


class LinearRunner(BaseRunner):
    """
    A thin wrapper around the NiPype LinearPlugin used to
    run pipelines on the local workstation on a single process

    Parameters
    ----------
    work_dir : str
        A directory in which to run the nipype workflows
    max_process_time : float
        The maximum time allowed for the process
    reprocess: True|False|'all'
        A flag which determines whether to rerun the processing for this
        step. If set to 'all' then pre-requisite pipelines will also be
        reprocessed.
    """

    nipype_plugin_cls = LinearPlugin

    num_processes = 1



from itertools import chain
from collections import OrderedDict
from arcana.exception import ArcanaNameError


class TreeNode(object):

    def __init__(self, datasets, fields):
        if datasets is None:
            datasets = []
        if fields is None:
            fields = []
        self._datasets = OrderedDict((d.name, d) for d in datasets)
        self._fields = OrderedDict((f.name, f) for f in fields)

    @property
    def datasets(self):
        return self._datasets.itervalues()

    @property
    def fields(self):
        return self._fields.itervalues()

    @property
    def dataset_names(self):
        return self._datasets.iterkeys()

    @property
    def field_names(self):
        return self._fields.iterkeys()

    def dataset(self, name):
        try:
            return self._datasets[name]
        except KeyError:
            raise ArcanaNameError(
                name, ("{} doesn't have a dataset named '{}' "
                       "(available '{}')"
                       .format(self, name,
                               "', '".join(self.dataset_names))))

    def field(self, name):
        try:
            return self._fields[name]
        except KeyError:
            raise ArcanaNameError(
                name, ("{} doesn't have a field named '{}' "
                       "(available '{}')"
                       .format(self, name,
                               "', '".join(self.field_names))))

    @property
    def data(self):
        return chain(self.datasets, self.fields)

    @property
    def data_names(self):
        return (d.name for d in self.data)

    def __eq__(self, other):
        if not (isinstance(other, type(self)) or
                isinstance(self, type(other))):
            return False
        return (self._datasets == other._datasets and
                self._fields == other._fields)

    def __ne__(self, other):
        return not (self == other)

    def find_mismatch(self, other, indent=''):
        if self != other:
            mismatch = "\n{}{}".format(indent, type(self).__name__)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if len(list(self.datasets)) != len(list(other.datasets)):
            mismatch += ('\n{indent}mismatching summary dataset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.datasets)),
                                 len(list(other.datasets)),
                                 list(self.datasets),
                                 list(other.datasets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.datasets, other.datasets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching summary field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch


class Project(TreeNode):
    """
    Represents a project tree as stored in an archive

    Parameters
    ----------
    subjects : List[Subject]
        List of subjects
    visits : List[Visits]
        List of visits in the project across subjects
        (i.e. timepoint 1, 2, 3)
    datasets : List[Dataset]
        The datasets that belong to the project, i.e. of 'per_project'
        frequency
    fields : List[Field]
        The fields that belong to the project, i.e. of 'per_project'
        frequency
    """

    def __init__(self, subjects, visits, datasets=None, fields=None):
        TreeNode.__init__(self, datasets, fields)
        self._subjects = {s.id: s for s in subjects}
        self._visits = {v.id: v for v in visits}

    @property
    def subjects(self):
        return self._subjects.itervalues()

    @property
    def visits(self):
        return self._visits.itervalues()

    def subject(self, id):  # @ReservedAssignment
        try:
            return self._subjects[id]
        except KeyError:
            raise ArcanaNameError(
                id, ("{} doesn't have a subject named '{}'"
                       .format(self, id)))

    def visit(self, id):  # @ReservedAssignment
        try:
            return self._visits[id]
        except KeyError:
            raise ArcanaNameError(
                id, ("{} doesn't have a visit named '{}'"
                       .format(self, id)))

    def nodes(self, frequency):
        if frequency == 'per_session':
            nodes = chain(*(s.sessions for s in self.subjects))
        elif frequency == 'per_subject':
            nodes = self.subjects
        elif frequency == 'per_visit':
            nodes = self.visits
        elif frequency == 'per_project':
            nodes = [self]
        else:
            assert False
        return nodes

    def __eq__(self, other):
        return (super(Project, self).__eq__(other) and
                self._subjects == other._subjects and
                self._visits == other._visits)

    def find_mismatch(self, other, indent=''):
        """
        Used in debugging unittests
        """
        mismatch = super(Project, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if len(list(self.subjects)) != len(list(other.subjects)):
            mismatch += ('\n{indent}mismatching subject lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.subjects)),
                                 len(list(other.subjects)),
                                 list(self.subjects),
                                 list(other.subjects),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.subjects, other.subjects):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.visits)) != len(list(other.visits)):
            mismatch += ('\n{indent}mismatching visit lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.visits)),
                                 len(list(other.visits)),
                                 list(self.visits),
                                 list(other.visits),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.visits, other.visits):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __repr__(self):
        return ("Project(num_subjects={}, num_visits={}, "
                "num_datasets={}, num_fields={})".format(
                    len(list(self.subjects)),
                    len(list(self.visits)),
                    len(list(self.datasets)), len(list(self.fields))))


class Subject(TreeNode):
    """
    Represents a subject as stored in an archive

    Parameters
    ----------
    subject_id : str
        The ID of the subject
    sessions : List[Session]
        The sessions in the subject
    datasets : List[Dataset]
        The datasets that belong to the subject, i.e. of 'per_subject'
        frequency
    fields : List[Field]
        The fields that belong to the subject, i.e. of 'per_subject'
        frequency
    """

    def __init__(self, subject_id, sessions, datasets=None,
                 fields=None):
        TreeNode.__init__(self, datasets, fields)
        self._id = subject_id
        self._sessions = {s.visit_id: s for s in sessions}
        for session in self.sessions:
            session.subject = self

    @property
    def id(self):
        return self._id

    def __lt__(self, other):
        return self._id < other._id

    @property
    def sessions(self):
        return self._sessions.itervalues()

    def session(self, visit_id):
        try:
            return self._sessions[visit_id]
        except KeyError:
            raise ArcanaNameError(
                visit_id, ("{} doesn't have a session named '{}'"
                           .format(self, visit_id)))

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other)and
                self._id == other._id and
                self._sessions == other._sessions)

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Subject(id={}, num_sessions={})"
                .format(self._id, len(self._sessions)))


class Visit(TreeNode):
    """
    Represents a collection of visits across subjects (e.g. time-point 1)
    as stored in an archive

    Parameters
    ----------
    visit_id : str
        The ID of the visit
    sessions : List[Session]
        The sessions in the visit
    datasets : List[Dataset]
        The datasets that belong to the visit, i.e. of 'per_visit'
        frequency
    fields : List[Field]
        The fields that belong to the visit, i.e. of 'per_visit'
        frequency
    """

    def __init__(self, visit_id, sessions, datasets=None, fields=None):
        TreeNode.__init__(self, datasets, fields)
        self._id = visit_id
        self._sessions = {s.subject_id: s for s in sessions}
        for session in sessions:
            session.visit = self

    @property
    def id(self):
        return self._id

    def __lt__(self, other):
        return self._id < other._id

    @property
    def sessions(self):
        return self._sessions.itervalues()

    def session(self, subject_id):
        try:
            return self._sessions[subject_id]
        except KeyError:
            raise ArcanaNameError(
                subject_id, ("{} doesn't have a session named '{}'"
                             .format(self, subject_id)))

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other)and
                self._id == other._id and
                self._sessions == other._sessions)

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "Visit(id={}, num_sessions={})".format(self._id,
                                                      len(self._sessions))


class Session(TreeNode):
    """
    Represents a session stored in an archive

    Parameters
    ----------
    subject_id : str
        The subject ID of the session
    visit_id : str
        The visit ID of the session
    datasets : list(Dataset)
        The datasets found in the session
    derived : Session
        If derived scans are stored in a separate session, it is provided
        here
    """

    def __init__(self, subject_id, visit_id, datasets=None, fields=None,
                 derived=None):
        TreeNode.__init__(self, datasets, fields)
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._subject = None
        self._visit = None
        self._derived = derived

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def subject_id(self):
        return self._subject_id

    def __lt__(self, other):
        if self.subject_id < other.subject_id:
            return True
        else:
            return self.visit_id < other.visit_id

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def visit(self):
        return self._visit

    @visit.setter
    def visit(self, visit):
        self._visit = visit

    @property
    def derived(self):
        return self._derived

    @derived.setter
    def derived(self, derived):
        self._derived = derived

    @property
    def acquired(self):
        """True if the session contains acquired scans"""
        return not self._derived or self._derived is None

    @property
    def derived_dataset_names(self):
        datasets = (self.datasets
                    if self.derived is None else self.derived.datasets)
        return (d.name for d in datasets)

    @property
    def derived_field_names(self):
        fields = (self.fields
                  if self.derived is None else self.derived.fields)
        return (f.name for f in fields)

    @property
    def derived_data_names(self):
        return chain(self.derived_dataset_names,
                     self.derived_field_names)

    @property
    def all_dataset_names(self):
        return chain(self.dataset_names, self.derived_dataset_names)

    @property
    def all_field_names(self):
        return chain(self.field_names, self.derived_field_names)

    @property
    def all_data_names(self):
        return chain(self.data_names, self.derived_data_names)

    def __eq__(self, other):
        return (TreeNode.__eq__(self, other) and
                self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self.derived == other.derived)

    def find_mismatch(self, other, indent=''):
        mismatch = TreeNode.find_mismatch(self, other, indent)
        sub_indent = indent + '  '
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.derived != other.derived:
            mismatch += ('\n{}derived: self={} v other={}'
                         .format(sub_indent, self.derived,
                                 other.derived))
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Session(subject_id='{}', visit_id='{}', num_datasets={}, "
                "num_fields={}, derived={})".format(
                    self.subject_id, self.visit_id, len(self._datasets),
                    len(self._fields), self.derived))



import os.path
import pydicom
from arcana.data_format import DataFormat, directory_format
from arcana.utils import split_extension
from arcana.exception import (
    ArcanaError, ArcanaDataFormatError, ArcanaUsageError,
    ArcanaDataFormatNotRegisteredError)
from .base import BaseDataset, BaseField


class Dataset(BaseDataset):
    """
    A representation of a dataset within the archive.

    Parameters
    ----------
    name : str
        The name of the dataset
    format : FileFormat
        The file format used to store the dataset. Can be one of the
        recognised formats
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_project',
        specifying whether the dataset is present for each session, subject,
        visit or project.
    derived : bool
        Whether the scan was generated or acquired. Depending on the archive
        used to store the dataset this is used to determine the location of the
        dataset.
    path : str | None
        The path to the dataset (for archives on the local system)
    id : int | None
        The ID of the dataset in the session. To be used to
        distinguish multiple datasets with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        archives where this isn't stored (i.e. Local), id can be None
    subject_id : int | str | None
        The id of the subject which the dataset belongs to
    visit_id : int | str | None
        The id of the visit which the dataset belongs to
    archive : BaseArchive
        The archive which the dataset is stored
    """

    is_spec = False

    def __init__(self, name, format=None, derived=False,  # @ReservedAssignment @IgnorePep8
                 frequency='per_session', path=None,
                 id=None, uri=None, subject_id=None, visit_id=None,  # @ReservedAssignment @IgnorePep8
                 archive=None):
        super(Dataset, self).__init__(name, format, frequency)
        self._derived = derived
        self._path = path
        self._id = id
        self._uri = uri
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._archive = archive

    def __eq__(self, other):
        return (super(Dataset, self).__eq__(other) and
                self.derived == other.derived and
                self._path == other._path and
                self.id == other.id and
                self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self._archive == other._archive)

    def __lt__(self, other):
        return self.id < other.id

    def find_mismatch(self, other, indent=''):
        mismatch = super(Dataset, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if self.derived != other.derived:
            mismatch += ('\n{}derived: self={} v other={}'
                         .format(sub_indent, self.derived,
                                 other.derived))
        if self._path != other._path:
            mismatch += ('\n{}path: self={} v other={}'
                         .format(sub_indent, self._path,
                                 other._path))
        if self._id != other._id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self._id,
                                 other._id))
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.archive != other.archive:
            mismatch += ('\n{}archive: self={} v other={}'
                         .format(sub_indent, self.archive,
                                 other.archive))
        return mismatch

    @property
    def path(self):
        if self._path is None:
            if self.archive is not None:
                self._path = self.archive.cache(self)
            else:
                raise ArcanaError(
                    "Neither path nor archive has been set for Dataset "
                    "{}".format(self.name))
        return self._path

    def basename(self, **kwargs):  # @UnusedVariable
        return self.name

    @property
    def derived(self):
        return self._derived

    @property
    def id(self):
        if self._id is None:
            return self.name
        else:
            return self._id

    @property
    def uri(self):
        return self._uri

    @property
    def archive(self):
        return self._archive

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def visit_id(self):
        return self._visit_id

    @classmethod
    def from_path(cls, path, frequency='per_session', format=None,  # @ReservedAssignment @IgnorePep8
                  subject_id=None, visit_id=None, archive=None):
        if not os.path.exists(path):
            raise ArcanaUsageError(
                "Attempting to read Dataset from path '{}' but it "
                "does not exist".format(path))
        if os.path.isdir(path):
            within_exts = frozenset(
                split_extension(f)[1] for f in os.listdir(path)
                if not f.startswith('.'))
            if format is None:
                # Try to guess format
                try:
                    format = DataFormat.by_within_dir_exts(within_exts)  # @ReservedAssignment @IgnorePep8
                except ArcanaDataFormatNotRegisteredError:
                    # Fall back to general directory format
                    format = directory_format  # @ReservedAssignment
            name = os.path.basename(path)
        else:
            filename = os.path.basename(path)
            name, ext = split_extension(filename)
            if format is None:
                try:
                    format = DataFormat.by_ext(ext)  # @ReservedAssignment @IgnorePep8
                except ArcanaDataFormatNotRegisteredError as e:
                    raise ArcanaDataFormatNotRegisteredError(
                        str(e) + ", which is required to identify the "
                        "format of the dataset at '{}'".format(path))
        return cls(name, format, frequency=frequency,
                   path=path, derived=False, subject_id=subject_id,
                   visit_id=visit_id, archive=archive)

    def dicom(self, index):
        """
        Returns a PyDicom object for the DICOM file at index 'index'

        Parameters
        ----------
        dataset : Dataset
            The dataset to read a DICOM file from
        index : int
            The index of the DICOM file in the dataset to read

        Returns
        -------
        dcm : pydicom.DICOM
            A PyDicom file object
        """
        if self.format.name != 'dicom':
            raise ArcanaDataFormatError(
                "Can not read DICOM header as {} is not in DICOM format"
                .format(self))
        fnames = sorted([f for f in os.listdir(self.path)
                         if not f.startswith('.')])
        with open(os.path.join(self.path, fnames[index])) as f:
            dcm = pydicom.dcmread(f)
        return dcm

    def dicom_values(self, tags, archive_login=None):
        """
        Returns a dictionary with the DICOM header fields corresponding
        to the given tag names

        Parameters
        ----------
        tags : List[Tuple[str, str]]
            List of DICOM tag values as 2-tuple of strings, e.g.
            [('0080', '0020')]
        archive_login : <archive-login-object>
            A login object for the archive to avoid having to relogin
            for every dicom_header call.

        Returns
        -------
        dct : Dict[Tuple[str, str], str|int|float]
        """
        if (self._path is None and self._archive is not None and
                hasattr(self.archive, 'dicom_header')):
            hdr = self.archive.dicom_header(self,
                                            prev_login=archive_login)
            dct = {t: hdr[t] for t in tags}
        else:
            # Get the DICOM object for the first file in the dataset
            dcm = self.dicom(0)
            dct = {t: dcm[t].value for t in tags}
        return dct

    def initkwargs(self):
        dct = super(Dataset, self).initkwargs()
        dct['derived'] = self.derived
        dct['path'] = self.path
        dct['id'] = self.id
        dct['uri'] = self.uri
        dct['archive'] = self.archive
        return dct


class Field(BaseField):
    """
    A representation of a value field in the archive.

    Parameters
    ----------
    name : str
        The name of the dataset
    dtype : type
        The datatype of the value. Can be one of (float, int, str)
    frequency : str
        One of 'per_session', 'per_subject', 'per_visit' and 'per_project',
        specifying whether the dataset is present for each session, subject,
        visit or project.
    derived : bool
        Whether or not the value belongs in the derived session or not
    subject_id : int | str | None
        The id of the subject which the field belongs to
    visit_id : int | str | None
        The id of the visit which the field belongs to
    archive : BaseArchive
        The archive which the field is stored
    """

    def __init__(self, name, value, frequency='per_session',
                 derived=False, subject_id=None, visit_id=None,
                 archive=None):
        if isinstance(value, int):
            dtype = int
        elif isinstance(value, float):
            dtype = float
        elif isinstance(value, basestring):
            # Attempt to implicitly convert from string
            try:
                value = int(value)
                dtype = int
            except ValueError:
                try:
                    value = float(value)
                    dtype = float
                except ValueError:
                    dtype = str
        else:
            raise ArcanaError(
                "Unrecognised field dtype {} (can be int, float or str)"
                .format(value))
        super(Field, self).__init__(
            name, dtype, frequency=frequency)
        self._value = value
        self._derived = derived
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._archive = archive

    def __eq__(self, other):
        return (super(Field, self).__eq__(other) and
                self.derived == other.derived and
                self.value == other.value and
                self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self._archive == other._archive)

    def find_mismatch(self, other, indent=''):
        mismatch = super(Field, self).find_mismatch(other, indent)
        sub_indent = indent + '  '
        if self.derived != other.derived:
            mismatch += ('\n{}derived: self={} v other={}'
                         .format(sub_indent, self.derived,
                                 other.derived))
        if self.value != other.value:
            mismatch += ('\n{}value: self={} v other={}'
                         .format(sub_indent, self.value,
                                 other.value))
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.archive != other.archive:
            mismatch += ('\n{}archive: self={} v other={}'
                         .format(sub_indent, self.archive,
                                 other.archive))
        return mismatch

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return ("{}(name='{}', value={}, frequency='{}', derived={},"
                " subject_id={}, visit_id={}, archive={})".format(
                    type(self).__name__, self.name, self.value,
                    self.frequency, self.derived, self.subject_id,
                    self.visit_id, self.archive))

    @property
    def derived(self):
        return self._derived

    def basename(self, **kwargs):  # @UnusedVariable
        return self.name

    @property
    def archive(self):
        return self._archive

    @property
    def value(self):
        return self._value

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def visit_id(self):
        return self._visit_id

    def initkwargs(self):
        dct = {}
        dct['name'] = self.name
        dct['value'] = self.value
        dct['frequency'] = self.frequency
        dct['derived'] = self.derived
        dct['archive'] = self.archive
        return dct
